"""Data handling scripts."""

from os import PathLike
from typing import Iterable, List

import click
from nibabel.nifti1 import Nifti1Header
from numpy import ndarray


def compute_default_filename(out_path):
    from os import path

    abspath = path.abspath(out_path)
    # If given path is a directory, append a filename
    if path.isdir(abspath):
        tsv_path = path.join(out_path, "merge.tsv")
    elif "." not in path.basename(abspath):
        tsv_path = f"{out_path}.tsv"
    else:
        if path.splitext(out_path)[1] != ".tsv":
            raise TypeError("Output path extension must be tsv.")
        tsv_path = out_path

    return tsv_path


def create_merge_file(
    bids_dir,
    out_tsv,
    caps_dir=None,
    tsv_file=None,
    pipelines=None,
    ignore_scan_files=False,
    ignore_sessions_files=False,
    **kwargs,
):
    """Merge all the TSV files containing clinical data of a BIDS compliant dataset and store the result inside a TSV file.

    Args:
        bids_dir: path to the BIDS folder
        out_tsv: path to the output tsv file
        caps_dir: path to the CAPS folder (optional)
        tsv_file: TSV file containing the subjects with their sessions (optional)
        ignore_scan_files: If True the information related to scans is not read (optional)
        ignore_sessions_files: If True the information related to sessions and scans is not read (optional)
        pipelines: when adding CAPS information, indicates the pipelines that will be merged (optional)
    """
    import json
    import os
    from os import path

    import numpy as np
    import pandas as pd

    from clinica.utils.participant import get_subject_session_list
    from clinica.utils.stream import cprint

    from .pipeline_handling import DatasetError

    if caps_dir is not None:
        if not path.isdir(caps_dir):
            raise IOError("The path to the CAPS directory is wrong")

    sessions, subjects = get_subject_session_list(
        bids_dir, ss_file=tsv_file, use_session_tsv=(not ignore_sessions_files)
    )

    if not os.path.isfile(path.join(bids_dir, "participants.tsv")):
        participants_df = pd.DataFrame(list(set(subjects)), columns=["participant_id"])
    else:
        participants_df = pd.read_csv(path.join(bids_dir, "participants.tsv"), sep="\t")

    sub_ses_df = pd.DataFrame(
        [[subject, session] for subject, session in zip(subjects, sessions)],
        columns=["participant_id", "session_id"],
    )

    try:
        sub_ses_df.set_index(
            ["participant_id", "session_id"], inplace=True, verify_integrity=True
        )
    except ValueError:
        cprint(
            "Found duplicate subject/session pair. Keeping first occurrence.",
            lvl="warning",
        )
        sub_ses_df = sub_ses_df.drop_duplicates(subset=["participant_id", "session_id"])
        sub_ses_df.set_index(["participant_id", "session_id"], inplace=True)

    out_path = compute_default_filename(out_tsv)
    out_dir = path.dirname(out_path)
    if len(out_dir) > 0:
        os.makedirs(out_dir, exist_ok=True)

    merged_df = pd.DataFrame(columns=participants_df.columns.values)

    # BIDS part
    for subject, subject_df in sub_ses_df.groupby(level=0):
        sub_path = path.join(bids_dir, subject)
        row_participant_df = participants_df[
            participants_df["participant_id"] == subject
        ]
        row_participant_df.reset_index(inplace=True, drop=True)
        if len(row_participant_df) == 0:
            cprint(
                msg=f"Participant {subject} does not exist in participants.tsv",
                lvl="warning",
            )
            row_participant_df = pd.DataFrame([[subject]], columns=["participant_id"])

        if ignore_sessions_files:
            for _, session in subject_df.index.values:
                row_session_df = pd.DataFrame([[session]], columns=["session_id"])

                row_df = pd.concat([row_participant_df, row_session_df], axis=1)
                merged_df = pd.concat([merged_df, row_df])

        else:
            sessions_df = pd.read_csv(
                path.join(sub_path, f"{subject}_sessions.tsv"), sep="\t"
            )

            for _, session in subject_df.index.values:
                row_session_df = sessions_df[sessions_df.session_id == session]
                row_session_df.reset_index(inplace=True, drop=True)
                if len(row_session_df) == 0:
                    raise DatasetError(
                        sessions_df.loc[0, "session_id"] + " / " + session
                    )

                # Read scans TSV files
                scan_path = path.join(
                    bids_dir,
                    subject,
                    session,
                    f"{subject}_{session}_scans.tsv",
                )
                if path.isfile(scan_path) and not ignore_scan_files:
                    scans_dict = dict()
                    scans_df = pd.read_csv(scan_path, sep="\t")
                    for idx in scans_df.index.values:
                        filepath = scans_df.loc[idx, "filename"]
                        if filepath.endswith(".nii.gz"):
                            filename = path.basename(filepath).split(".")[0]
                            modality = "_".join(filename.split("_")[2::])
                            for col in scans_df.columns.values:
                                if col == "filename":
                                    pass
                                else:
                                    value = scans_df.loc[idx, col]
                                    new_col_name = f"{modality}_{col}"
                                    scans_dict.update({new_col_name: value})
                            json_path = path.join(
                                bids_dir,
                                subject,
                                session,
                                filepath.split(".")[0] + ".json",
                            )
                            if path.exists(json_path):
                                with open(json_path, "r") as f:
                                    json_dict = json.load(f)
                                for key, value in json_dict.items():
                                    new_col_name = f"{modality}_{key}"
                                    scans_dict.update({new_col_name: value})
                    scans_dict = {
                        str(key): str(value) for key, value in scans_dict.items()
                    }
                    row_scans_df = pd.DataFrame(scans_dict, index=[0])
                else:
                    row_scans_df = pd.DataFrame()

                row_df = pd.concat(
                    [row_participant_df, row_session_df, row_scans_df], axis=1
                )

                # remove duplicated columns
                row_df = row_df.loc[:, ~row_df.columns.duplicated(keep="last")]

                merged_df = pd.concat([merged_df, row_df])

    # Put participant_id and session_id first
    col_list = merged_df.columns.values.tolist()
    col_list.insert(0, col_list.pop(col_list.index("participant_id")))
    col_list.insert(1, col_list.pop(col_list.index("session_id")))
    merged_df = merged_df[col_list]

    tmp = merged_df.select_dtypes(include=[np.number])
    # Round numeric values in dataframe to 6 decimal values
    merged_df.loc[:, tmp.columns] = np.round(tmp, 6)
    merged_df.to_csv(out_path, sep="\t", index=False)
    cprint("End of BIDS information merge.", lvl="debug")

    merged_df.reset_index(drop=True, inplace=True)

    # CAPS
    if caps_dir is not None:
        # Call the different pipelines
        from .pipeline_handling import (
            dwi_dti_pipeline,
            pet_volume_pipeline,
            t1_freesurfer_longitudinal_pipeline,
            t1_freesurfer_pipeline,
            t1_volume_pipeline,
        )

        pipeline_options = {
            "t1-volume": t1_volume_pipeline,
            "pet-volume": pet_volume_pipeline,
            "t1-freesurfer": t1_freesurfer_pipeline,
            "t1-freesurfer-longitudinal": t1_freesurfer_longitudinal_pipeline,
            "dwi-dti": dwi_dti_pipeline,
        }
        merged_summary_df = pd.DataFrame()
        if not pipelines:
            for pipeline_name, pipeline_fn in pipeline_options.items():
                merged_df, summary_df = pipeline_fn(caps_dir, merged_df, **kwargs)
                if summary_df is not None and not summary_df.empty:
                    merged_summary_df = pd.concat([merged_summary_df, summary_df])

                if summary_df is None or summary_df.empty:
                    cprint(
                        f"{pipeline_name} outputs were not found in the CAPS folder."
                    )
        else:
            for pipeline in pipelines:
                merged_df, summary_df = pipeline_options[pipeline](
                    caps_dir, merged_df, **kwargs
                )
                merged_summary_df = pd.concat([merged_summary_df, summary_df])

        n_atlas = len(merged_summary_df)
        if n_atlas == 0:
            raise FileNotFoundError(
                "No outputs were found for any pipeline in the CAPS folder. "
                "The output only contains BIDS information."
            )
        columns = merged_df.columns.values.tolist()
        merged_summary_df.reset_index(inplace=True, drop=True)
        for idx in merged_summary_df.index:
            first_column_name = merged_summary_df.loc[idx, "first_column_name"]
            last_column_name = merged_summary_df.loc[idx, "last_column_name"]
            merged_summary_df.loc[idx, "first_column_index"] = columns.index(
                first_column_name
            )
            merged_summary_df.loc[idx, "last_column_index"] = columns.index(
                last_column_name
            )

        summary_path = path.splitext(out_path)[0] + "_summary.tsv"
        merged_summary_df.to_csv(summary_path, sep="\t", index=False)

        tmp = merged_df.select_dtypes(include=[np.number])
        # Round numeric values in dataframe to 12 floating point values
        merged_df.loc[:, tmp.columns] = np.round(tmp, 12)
        merged_df.to_csv(out_path, sep="\t", index=False)
        cprint("End of CAPS information merge.", lvl="debug")


def find_mods_and_sess(bids_dir):
    """Find all the modalities and sessions available for a given BIDS dataset.

    Args:
        bids_dir: path to the BIDS dataset

    Returns:
        mods_dict: a dictionary that stores the sessions and modalities found and has the following structure.
    Example:
    {
        'sessions': ['ses-M00', 'ses-M18'],
        'fmap': ['fmap'],
        'anat': ['flair', 't1w'],
        'func': ['func_task-rest'],
        'dwi': ['dwi']
    }

    """
    import os
    from glob import glob
    from os import path

    mods_dict = {}
    mods_list = []
    subjects_paths_lists = glob(path.join(bids_dir, "*sub-*"))

    for sub_path in subjects_paths_lists:
        ses_paths = glob(path.join(sub_path, "*ses-*"))
        for session in ses_paths:
            ses_name = session.split(os.sep)[-1]
            mods_avail = []
            if "sessions" in mods_dict:
                if ses_name not in mods_dict["sessions"]:
                    mods_dict["sessions"].append(ses_name)
            else:
                mods_dict.update({"sessions": [ses_name]})
            mods_paths_folders = glob(path.join(session, "*/"))

            for p in mods_paths_folders:
                p = p[:-1]
                mods_avail.append(p.split("/").pop())
            if "func" in mods_avail:
                list_funcs_paths = glob(path.join(session, "func", "*bold.nii.gz"))
                for func_path in list_funcs_paths:
                    func_name = func_path.split(os.sep)[-1]
                    func_name_tokens = func_name.split("_")
                    func_task = func_name_tokens[2]
                    if "func" in mods_dict:
                        if "func_" + func_task not in mods_dict["func"]:
                            mods_dict["func"].append("func_" + func_task)
                    else:
                        mods_dict.update({"func": ["func_" + func_task]})

                    if "func_" + func_task not in mods_list:
                        mods_list.append("func_" + func_task)

            if "dwi" in mods_avail:
                if "dwi" not in mods_dict:
                    mods_dict.update({"dwi": ["dwi"]})
                if "dwi" not in mods_list:
                    mods_list.append("dwi")

            if "fmap" in mods_avail:
                if "fmap" not in mods_dict:
                    mods_dict.update({"fmap": ["fmap"]})
                if "fmap" not in mods_list:
                    mods_list.append("fmap")

            if "pet" in mods_avail:
                list_pet_paths = glob(path.join(session, "pet", "*pet.nii.gz"))
                for pet_path in list_pet_paths:
                    pet_name = pet_path.split(os.sep)[-1].split(".")[0]
                    pet_name_tokens = pet_name.split("_")
                    # FIXME: Use an appropriate parser for entity extraction.
                    pet_acq = pet_name_tokens[2]
                    if "pet" in mods_dict:
                        if "pet_" + pet_acq not in mods_dict["pet"]:
                            mods_dict["pet"].append("pet_" + pet_acq)
                    else:
                        mods_dict.update({"pet": ["pet_" + pet_acq]})

                    if "pet_" + pet_acq not in mods_list:
                        mods_list.append("pet_" + pet_acq)

            if "anat" in mods_avail:
                anat_files_paths = glob(path.join(session, "anat", "*"))

                for anat_file in anat_files_paths:
                    anat_name = anat_file.split(os.sep)[-1]

                    # Extract the name of the file without the extension
                    if ".nii.gz" in anat_name:
                        anat_name = anat_name.replace(".nii.gz", "")
                        anat_ext = "nii.gz"
                    else:
                        anat_name = os.path.splitext(anat_name.split(os.sep)[-1])[0]
                        anat_ext = os.path.splitext(anat_name.split(os.sep)[-1])[1]

                    if anat_ext != "json":
                        file_parts = anat_name.split("_")
                        anat_type = str.lower(file_parts[len(file_parts) - 1])
                        if "anat" in mods_dict:
                            if anat_type not in mods_dict["anat"]:
                                anat_aval = mods_dict["anat"]
                                anat_aval.append(anat_type)
                                mods_dict.update({"anat": anat_aval})
                        else:
                            mods_dict.update({"anat": [anat_type]})

                        if anat_type not in mods_list:
                            mods_list.append(anat_type)

    return mods_dict


def compute_missing_processing(bids_dir, caps_dir, out_file):
    """
    Compute the list of missing processing for each subject in a CAPS compliant dataset

    Args:
        bids_dir: path to the BIDS directory.
        caps_dir: path to the CAPS directory.
        out_file: path to the output file (filename included).
    """
    from glob import glob
    from os import listdir, path, sep

    import pandas as pd

    if path.exists(path.join(caps_dir, "groups")):
        groups = listdir(path.join(caps_dir, "groups"))
    else:
        groups = list()
    output_df = pd.DataFrame()

    # Retrieve pet tracers avail
    mods_and_sess = find_mods_and_sess(bids_dir)
    mods_and_sess.pop("sessions")
    mods_avail_dict = mods_and_sess
    trc_avail = [
        j.split("_")[1] for i in mods_avail_dict.values() for j in i if "pet" in j
    ]
    print(trc_avail)

    subjects_paths = glob(path.join(caps_dir, "subjects", "sub-*"))
    for subject_path in subjects_paths:
        participant_id = subject_path.split(sep)[-1]
        sessions_paths = glob(path.join(subject_path, "ses-*"))
        for session_path in sessions_paths:
            session_id = session_path.split(sep)[-1]
            row_df = pd.DataFrame(
                [[participant_id, session_id]], columns=["participant_id", "session_id"]
            )

            # Check t1-volume outputs
            if path.exists(path.join(session_path, "t1", "spm", "segmentation")):
                row_df.loc[0, "t1-volume-segmentation"] = "1"
                for group in groups:
                    group_id = group.split("-")[-1]
                    if path.exists(
                        path.join(
                            session_path,
                            "t1",
                            "spm",
                            "dartel",
                            group,
                            f"{participant_id}_{session_id}_T1w_target-{group_id}_transformation-forward_deformation.nii.gz",
                        )
                    ):
                        row_df.loc[0, f"t1-volume-register-dartel_{group}"] = "1"
                        dartel2mni = glob(
                            path.join(
                                session_path,
                                "t1",
                                "spm",
                                "dartel",
                                group,
                                f"{participant_id}_{session_id}_T1w_segm-*_space-Ixi549Space_modulated-*_probability.nii.gz",
                            )
                        )
                        if len(dartel2mni) > 0:
                            row_df.loc[0, f"t1-volume-dartel2mni_{group}"] = "1"
                            if path.exists(
                                path.join(
                                    session_path,
                                    "t1",
                                    "spm",
                                    "dartel",
                                    group,
                                    "atlas_statistics",
                                )
                            ):
                                row_df.loc[0, f"t1-volume-parcellation_{group}"] = "1"
                            else:
                                row_df.loc[0, f"t1-volume-parcellation_{group}"] = "0"
                        else:
                            row_df.loc[0, f"t1-volume-dartel2mni_{group}"] = "0"
                            row_df.loc[0, f"t1-volume-parcellation_{group}"] = "0"
                    else:
                        row_df.loc[0, f"t1-volume-register-dartel_{group}"] = "0"
                        row_df.loc[0, f"t1-volume-dartel2mni_{group}"] = "0"
                        row_df.loc[0, f"t1-volume-parcellation_{group}"] = "0"
            else:
                row_df.loc[0, "t1-volume-segmentation"] = "0"
                for group in groups:
                    row_df.loc[0, f"t1-volume-register-dartel_{group}"] = "0"
                    row_df.loc[0, f"t1-volume-dartel2mni_{group}"] = "0"
                    row_df.loc[0, f"t1-volume-parcellation_{group}"] = "0"

            # Check t1-linear outputs
            if path.exists(path.join(session_path, "t1_linear")):
                row_df.loc[0, "t1-linear"] = "1"
            else:
                row_df.loc[0, "t1-linear"] = "0"

            # Check t1-freesurfer outputs
            if path.exists(path.join(session_path, "t1", "freesurfer_cross_sectional")):
                row_df.loc[0, "t1-freesurfer"] = "1"
            else:
                row_df.loc[0, "t1-freesurfer"] = "0"

            # Check pet-volume outputs
            for group in groups:
                for trc in trc_avail:
                    for pvc in [True, False]:
                        pet_pattern = path.join(
                            session_path, "pet", "preprocessing", group, f"*{trc}*"
                        )
                        pet_paths = glob(pet_pattern)
                        if pvc:
                            pet_paths = [
                                pet_path for pet_path in pet_paths if "pvc" in pet_path
                            ]
                        else:
                            pet_paths = [
                                pet_path
                                for pet_path in pet_paths
                                if "pvc" not in pet_path
                            ]

                        if len(pet_paths) > 0:
                            row_df.loc[0, f"pet-volume_{trc}_{group}_pvc-{pvc}"] = "1"
                        else:
                            row_df.loc[0, f"pet-volume_{trc}_{group}_pvc-{pvc}"] = "0"

            # Check pet-surface outputs
            for trc in trc_avail:
                pet_pattern = path.join(session_path, "pet", "surface", f"*{trc}*")
                if len(glob(pet_pattern)) > 0:
                    row_df.loc[0, f"pet-surface_{trc}"] = "1"
                else:
                    row_df.loc[0, f"pet-surface_{trc}"] = "0"

            # Check pet-linear outputs
            for trc in trc_avail:
                pet_pattern = path.join(session_path, "pet_linear", f"*{trc}*")
                if len(glob(pet_pattern)) > 0:
                    row_df.loc[0, f"pet-linear_{trc}"] = "1"
                else:
                    row_df.loc[0, f"pet-linear_{trc}"] = "0"

            output_df = pd.concat([output_df, row_df])

    output_df.sort_values(["participant_id", "session_id"], inplace=True)
    output_df.to_csv(out_file, sep="\t", index=False)


def compute_missing_mods(bids_dir, out_dir, output_prefix=""):
    """Compute the list of missing modalities for each subject in a BIDS compliant dataset.

    Args:
        bids_dir: path to the BIDS directory
        out_dir: path to the output folder
        output_prefix: string that replace the default prefix ('missing_mods_') in the name of all the output files
    created
    """
    import os
    from glob import glob
    from os import path

    import pandas as pd

    from ..converter_utils import (
        MissingModsTracker,
        print_longitudinal_analysis,
        print_statistics,
    )

    os.makedirs(out_dir, exist_ok=True)

    # Find all the modalities and sessions available for the input dataset
    mods_and_sess = find_mods_and_sess(bids_dir)
    sessions_found = mods_and_sess["sessions"]
    mods_and_sess.pop("sessions")
    mods_avail_dict = mods_and_sess
    mods_avail = [j for i in mods_avail_dict.values() for j in i]
    cols_dataframe = mods_avail[:]
    cols_dataframe.insert(0, "participant_id")
    mmt = MissingModsTracker(sessions_found, mods_avail)

    if output_prefix == "":
        out_file_name = "missing_mods_"
    else:
        out_file_name = output_prefix + "_"

    summary_file = open(path.join(out_dir, out_file_name + "summary.txt"), "w")
    analysis_file = open(path.join(out_dir, "analysis.txt"), "w")
    missing_mods_df = pd.DataFrame(columns=cols_dataframe)
    row_to_append_df = pd.DataFrame(columns=cols_dataframe)
    subjects_paths_lists = glob(path.join(bids_dir, "*sub-*"))
    subjects_paths_lists.sort()

    if len(subjects_paths_lists) == 0:
        raise IOError("No subjects found or dataset not BIDS compliant.")
    # Check the modalities available for each session
    for ses in sessions_found:
        for sub_path in subjects_paths_lists:
            mods_avail_bids = []
            subj_id = sub_path.split(os.sep)[-1]
            row_to_append_df["participant_id"] = pd.Series(subj_id)
            ses_path_avail = glob(path.join(sub_path, ses))
            if len(ses_path_avail) == 0:
                mmt.increase_missing_ses(ses)
                for mod in mods_avail:
                    row_to_append_df[mod] = pd.Series("0")
            else:

                ses_path = ses_path_avail[0]
                mods_paths_folders = glob(path.join(ses_path, "*/"))

                for p in mods_paths_folders:
                    p = p[:-1]
                    mods_avail_bids.append(p.split("/").pop())

                # Check if a modality folder is available and if is empty
                if "func" in mods_avail_bids:
                    # Extract all the task available
                    for m in mods_avail_dict["func"]:
                        tokens = m.split("_")
                        task_name = tokens[1]
                        task_avail_list = glob(
                            path.join(ses_path, "func", "*" + task_name + "*")
                        )

                        if len(task_avail_list) == 0:
                            row_to_append_df[m] = pd.Series("0")
                        else:
                            row_to_append_df[m] = pd.Series("1")
                # If the folder is not available but the modality is
                # in the list of the available one mark it as missing
                else:
                    if "func" in mods_avail_dict:
                        for m in mods_avail_dict["func"]:
                            row_to_append_df[m] = pd.Series("0")
                            mmt.add_missing_mod(ses, m)

                if "dwi" in mods_avail_bids:
                    row_to_append_df["dwi"] = pd.Series("1")
                else:
                    if "dwi" in mods_avail:
                        row_to_append_df["dwi"] = pd.Series("0")
                        mmt.add_missing_mod(ses, "dwi")

                if "anat" in mods_avail_bids:
                    for m in mods_avail_dict["anat"]:
                        anat_aval_list = glob(path.join(ses_path, "anat", "*.nii.gz"))
                        anat_aval_list = [
                            elem for elem in anat_aval_list if m.lower() in elem.lower()
                        ]
                        if len(anat_aval_list) > 0:
                            row_to_append_df[m] = pd.Series("1")
                        else:
                            row_to_append_df[m] = pd.Series("0")
                            mmt.add_missing_mod(ses, m)
                else:
                    if "anat" in mods_avail_dict:
                        for m in mods_avail_dict["anat"]:
                            row_to_append_df[m] = pd.Series("0")
                            mmt.add_missing_mod(ses, m)

                if "fmap" in mods_avail_bids:
                    row_to_append_df["fmap"] = pd.Series("1")
                else:
                    if "fmap" in mods_avail:
                        row_to_append_df["fmap"] = pd.Series("0")
                        mmt.add_missing_mod(ses, "fmap")
                if "pet" in mods_avail_bids:
                    # Extract all the task available
                    for m in mods_avail_dict["pet"]:
                        tokens = m.split("_")
                        pet_acq = tokens[1]
                        acq_avail_list = glob(
                            path.join(ses_path, "pet", "*" + pet_acq + "*")
                        )

                        if len(acq_avail_list) == 0:
                            row_to_append_df[m] = pd.Series("0")
                        else:
                            row_to_append_df[m] = pd.Series("1")
                # If the folder is not available but the modality is
                # in the list of the available one mark it as missing
                else:
                    if "pet" in mods_avail_dict:
                        for m in mods_avail_dict["pet"]:
                            row_to_append_df[m] = pd.Series("0")
                            mmt.add_missing_mod(ses, m)

            missing_mods_df = pd.concat([missing_mods_df, row_to_append_df])
            row_to_append_df = pd.DataFrame(columns=cols_dataframe)

        missing_mods_df = missing_mods_df[cols_dataframe]
        missing_mods_df.to_csv(
            path.join(out_dir, out_file_name + ses + ".tsv"),
            sep="\t",
            index=False,
            encoding="utf-8",
        )
        missing_mods_df = pd.DataFrame(columns=cols_dataframe)

    print_statistics(summary_file, len(subjects_paths_lists), sessions_found, mmt)
    print_longitudinal_analysis(
        analysis_file, bids_dir, out_dir, sessions_found, out_file_name
    )


def create_subs_sess_list(
    input_dir, output_dir, file_name=None, is_bids_dir=True, use_session_tsv=False
):
    """Create the file subject_session_list.tsv that contains the list of the visits for each subject for a BIDS or CAPS compliant dataset.

    Args:
        input_dir (str): Path to the BIDS or CAPS directory.
        output_dir (str): Path to the output directory
        file_name: name of the output file
        is_bids_dir (boolean): Specify if input_dir is a BIDS directory or
            not (i.e. a CAPS directory)
        use_session_tsv (boolean): Specify if the list uses the sessions listed in the sessions.tsv files
    """
    import os
    from glob import glob
    from os import path

    import pandas as pd

    os.makedirs(output_dir, exist_ok=True)

    if not file_name:
        file_name = "subjects_sessions_list.tsv"
    subjs_sess_tsv = open(path.join(output_dir, file_name), "w")
    subjs_sess_tsv.write("participant_id" + "\t" + "session_id" + "\n")

    if is_bids_dir:
        path_to_search = input_dir
    else:
        path_to_search = path.join(input_dir, "subjects")
    subjects_paths = glob(path.join(path_to_search, "*sub-*"))

    # Sort the subjects list
    subjects_paths.sort()

    if len(subjects_paths) == 0:
        raise IOError("Dataset empty or not BIDS/CAPS compliant.")

    for sub_path in subjects_paths:
        subj_id = sub_path.split(os.sep)[-1]

        if use_session_tsv:
            session_df = pd.read_csv(
                path.join(sub_path, subj_id + "_sessions.tsv"), sep="\t"
            )
            session_df.dropna(how="all", inplace=True)
            session_list = list(session_df["session_id"].to_numpy())
            for session in session_list:
                subjs_sess_tsv.write(subj_id + "\t" + session + "\n")

        else:
            sess_list = glob(path.join(sub_path, "*ses-*"))

            for ses_path in sess_list:
                session_name = ses_path.split(os.sep)[-1]
                subjs_sess_tsv.write(subj_id + "\t" + session_name + "\n")

    subjs_sess_tsv.close()


def center_nifti_origin(input_image, output_image):
    """Put the origin of the coordinate system at the center of the image.

    Args:
        input_image: path to the input image
        output_image: path to the output image (where the result will be stored)

    Returns:
        path of the output image created
    """
    import os
    from os.path import isfile

    import nibabel as nib
    import numpy as np
    from nibabel.spatialimages import ImageFileError

    error_str = None
    try:
        img = nib.load(input_image)
    except FileNotFoundError:
        error_str = f"No such file {input_image}"
    except ImageFileError:
        error_str = f"File {input_image} could not be read"

    except Exception as e:
        error_str = f"File {input_image} could not be loaded with nibabel: {e}"

    if not error_str:
        try:
            canonical_img = nib.as_closest_canonical(img)
            hd = canonical_img.header

            qform = np.zeros((4, 4))
            for i in range(1, 4):
                qform[i - 1, i - 1] = hd["pixdim"][i]
                qform[i - 1, 3] = -1.0 * hd["pixdim"][i] * hd["dim"][i] / 2.0
            new_img = nib.Nifti1Image(
                canonical_img.get_data(caching="unchanged"), affine=qform, header=hd
            )

            # Without deleting already-existing file, nib.save causes a severe bug on Linux system
            if isfile(output_image):
                os.remove(output_image)

            nib.save(new_img, output_image)
            if not isfile(output_image):
                error_str = (
                    f"NIfTI file created but Clinica could not save it to {output_image}. "
                    "Please check that the output folder has the correct permissions."
                )
        except Exception as e:
            error_str = (
                "File "
                + input_image
                + " could not be processed with nibabel: "
                + str(e)
            )

    return output_image, error_str


def center_all_nifti(
    bids_dir: str,
    output_dir: str,
    modalities: Iterable[str] = None,
    center_all_files: bool = False,
) -> List[str]:
    """Center all the NIfTI images of the input BIDS folder into the empty output_dir specified in argument.

    All the files from bids_dir are copied into output_dir, then all the NIfTI images found are replaced by their
    centered version if their center is off the origin by more than 50 mm.

    Parameters
    ----------
    bids_dir: str
        Path to the BIDS directory.
    output_dir: str
        Path to the output directory where the centered files will be written to.
    modalities: iterable of str, optional
        Process these modalities only. Process all modalities otherwise.
    center_all_files: bool, default=False
        Center files that may cause problem for SPM if set to False, all files otherwise.

    Returns
    -------
    list of str
        Centered NIfTI files.
    """
    from glob import glob
    from os import listdir
    from os.path import basename, isdir, isfile, join
    from shutil import copy, copytree

    from clinica.utils.exceptions import ClinicaBIDSError
    from clinica.utils.inputs import check_bids_folder

    # output and input must be different, so that we do not mess with user's data
    if bids_dir == output_dir:
        raise ClinicaBIDSError("Input BIDS and output directories must be different")

    # check that input is a BIDS dir
    check_bids_folder(bids_dir)

    for f in listdir(bids_dir):
        if isdir(join(bids_dir, f)) and not isdir(join(output_dir, f)):
            copytree(join(bids_dir, f), join(output_dir, f), copy_function=copy)
        elif isfile(join(bids_dir, f)) and not isfile(join(output_dir, f)):
            copy(join(bids_dir, f), output_dir)

    pattern = join(output_dir, "**/*.nii*")
    nifti_files = glob(pattern, recursive=True)

    # Now filter this list by elements in modality list
    #   For each file:
    #       if any modality name (lowercase) is found in the basename of the file:
    #           keep the file
    nifti_files_filtered = (
        [
            f
            for f in nifti_files
            if any(elem.lower() in basename(f).lower() for elem in modalities)
        ]
        if modalities
        else nifti_files
    )

    # Remove those who are centered
    if not center_all_files:
        nifti_files_filtered = [
            file for file in nifti_files_filtered if not is_centered(file)
        ]

    all_errors = []
    for f in nifti_files_filtered:
        print("Handling " + f)
        _, current_error = center_nifti_origin(f, f)
        if current_error:
            all_errors.append(current_error)
    if len(all_errors) > 0:
        final_error_msg = (
            "Clinica encountered "
            + str(len(all_errors))
            + " error(s) while trying to center all NIfTI images.\n"
        )
        for error in all_errors:
            final_error_msg += "\n" + error
        raise RuntimeError(final_error_msg)
    return nifti_files_filtered


def write_list_of_files(file_list, output_file):
    """Save `file_list` list of files into `output_file` text file.

    Args:
        file_list: (list of str) of path to files
        output_file: (str) path to the output txt file

    Returns:
        output_file
    """
    from os.path import isfile

    assert isinstance(file_list, list), "First argument must be a list"
    assert isinstance(output_file, str), "Second argument must be a str"
    if isfile(output_file):
        return None

    text_file = open(output_file, "w+")
    for created_file in file_list:
        text_file.write(created_file + "\n")
    text_file.close()
    return output_file


def check_relative_volume_location_in_world_coordinate_system(
    label_1,
    nifti_list1,
    label_2,
    nifti_list2,
    bids_dir,
    modality,
    skip_question=False,
):
    """
    Check if the NIfTI file list nifti_list1 and nifti_list2 provided in argument are not too far apart, otherwise coreg
    in SPM may fail. Norm between center of volumes of 2 files must be less than 80 mm.

    Args:
        label_1: label of the first nifti_list1 files (used in potential warning message)
        nifti_list1: first set of files
        label_2: label of the second nifti_list
        nifti_list2: second set of files, must be same length as nifti_list1
        bids_dir: bids directory (used in potential warning message)
        modality: string that must be used in argument of: clinica iotools bids --modality <MODALITY> (used in potential
                warning message)
        skip_question: bool, optional
            disable prompts for user input (default is False)
    """
    import sys
    from os.path import abspath, basename

    import numpy as np

    from clinica.utils.stream import cprint

    center_coordinate_1 = [get_world_coordinate_of_center(file) for file in nifti_list1]
    center_coordinate_2 = [get_world_coordinate_of_center(file) for file in nifti_list2]

    l2_norm = [
        np.linalg.norm(center_1 - center_2)
        for center_1, center_2 in zip(center_coordinate_1, center_coordinate_2)
    ]
    pairs_with_problems = [i for i, norm in enumerate(l2_norm) if norm > 80]

    if len(pairs_with_problems) > 0:
        warning_message = (
            f"It appears that {str(len(pairs_with_problems))} "
            "pairs of files have an important relative offset. "
            "SPM co-registration has a high probability to fail on these files:\n\n"
        )

        # File column width : 3 spaces more than the longest string to display
        file1_width = max(
            3 + len(label_1),
            3
            + max(
                len(basename(file))
                for file in [nifti_list1[k] for k in pairs_with_problems]
            ),
        )
        file2_width = max(
            3 + len(label_2),
            3
            + max(
                len(basename(file))
                for file in [nifti_list2[k] for k in pairs_with_problems]
            ),
        )

        norm_width = len("Relative distance")

        warning_message += (
            "%-"
            + str(file1_width)
            + "s%-"
            + str(file2_width)
            + "s%-"
            + str(norm_width)
            + "s"
        ) % (label_1, label_2, "Relative distance")

        warning_message += "\n" + "-" * (file1_width + file2_width + norm_width) + "\n"
        for file1, file2, norm in zip(
            [nifti_list1[k] for k in pairs_with_problems],
            [nifti_list2[k] for k in pairs_with_problems],
            [l2_norm[k] for k in pairs_with_problems],
        ):
            # Nice formatting as array
            # % escape character
            # - aligned to the left, with the size of the column
            # s = string, f = float
            # . for precision with float
            # https://docs.python.org/2/library/stdtypes.html#string-formatting for more information
            warning_message += (
                "%-"
                + str(file1_width)
                + "s%-"
                + str(file2_width)
                + "s%-"
                + str(norm_width)
                + ".2f\n"
            ) % (str(basename(file1)), str(basename(file2)), norm)
        warning_message += (
            "\nClinica provides a tool to counter this problem by replacing the center "
            "of the volume at the origin of the world coordinates.\nUse the following "
            "command line to correct the header of the faulty NIFTI volumes in a new folder:\n\n"
            f"`clinica iotools center-nifti {abspath(bids_dir)} {abspath(bids_dir)}_centered --modality {modality}`\n\n"
            "You will find more information on the command by typing `clinica iotools center-nifti` in the console."
        )
        cprint(msg=warning_message, lvl="warning")

        if not skip_question:
            if not click.confirm("Do you still want to launch the pipeline?"):
                click.echo("Clinica will now exit...")
                sys.exit(0)


def check_volume_location_in_world_coordinate_system(
    nifti_list: list,
    bids_dir: PathLike,
    modality: str = "t1w",
    skip_question: bool = False,
) -> bool:
    """Check if images are centered around the origin of the world coordinate
    Parameters
    ----
    nifti_list: list
        list of path to nifti files or path
    bids_dir: PathLike
        path to bids directory associated with this check
    modality: str, optional
        the modality of the image. Default='t1w'.
    skip_question: bool, optional
        if True, assume answer is yes. Default=False.
    Returns
    -------
    bool
        True if they are centered, False otherwise
    Warns
    ------
    If volume is not centered on origin of the world coordinate system
    Notes
    -----
    the NIfTI file list provided in argument are approximately centered around the origin of the
    world coordinates. Otherwise, issues may arise with further processing such as SPM segmentation. When not centered,
    we warn the user of the problem propose to exit clinica to run clinica iotools center-nifti or to continue with the execution
    of the pipeline
    """

    import sys
    from os.path import abspath, basename

    import click
    import numpy as np

    flag_centered = True
    list_non_centered_files = [file for file in nifti_list if not is_centered(file)]
    if len(list_non_centered_files) > 0:
        centers = [
            get_world_coordinate_of_center(file) for file in list_non_centered_files
        ]
        l2_norm = [np.linalg.norm(center, ord=2) for center in centers]

        # File column width : 3 spaces more than the longest string to display
        file_width = 3 + max(len(basename(file)) for file in list_non_centered_files)

        # Center column width (with a fixed minimum size) : 3 spaces more than the longest string to display
        center_width = max(
            len("Coordinate of center") + 3,
            3 + max(len(str(center)) for center in centers),
        )

        warning_message = (
            f"It appears that {str(len(list_non_centered_files))} files "
            "have a center way out of the origin of the world coordinate system. SPM has a high "
            "probability to fail on these files (for co-registration or segmentation):\n\n"
        )
        warning_message += (
            "%-" + str(file_width) + "s%-" + str(center_width) + "s%-s"
        ) % ("File", "Coordinate of center", "Distance to origin")
        # 18 is the length of the string 'Distance to origin'
        warning_message += "\n" + "-" * (file_width + center_width + 18) + "\n"
        for file, center, l2 in zip(list_non_centered_files, centers, l2_norm):
            warning_message += (
                "%-" + str(file_width) + "s%-" + str(center_width) + "s%-25.2f\n"
            ) % (basename(file), str(center), l2)

        cmd_line = f"`clinica iotools center-nifti {abspath(bids_dir)} {abspath(bids_dir)}_centered --modality {modality}`"

        warning_message += (
            "\nIf you are trying to launch the t1-freesurfer pipeline, you can ignore this message "
            "if you do not want to run the pet-surface pipeline afterward."
        )

        warning_message += (
            "\nClinica provides a tool to counter this problem by replacing the center of the volume"
            " at the origin of the world coordinates.\nUse the following command line to correct the "
            f"header of the faulty NIFTI volumes in a new folder:\n{cmd_line}"
            "You will find more information on the command by typing "
            "clinica iotools center-nifti in the console."
        )

        click.echo(warning_message)

        flag_centered = False

        if not skip_question:
            if not click.confirm("Do you still want to launch the pipeline?"):
                click.echo("Clinica will now exit...")
                sys.exit(0)

    return flag_centered


def is_centered(nii_volume: PathLike, threshold_l2: int = 50) -> bool:
    """Checks if a NIfTI volume is centered on the origin of the world coordinate system.

    Parameters
    ---------
    nii_volume : PathLike
        path to NIfTI volume
    threshold_l2: int, optional
        Maximum distance between origin of the world coordinate system and the center of the volume to
        be considered centered. The threshold were SPM segmentation stops working is around 100 mm (it was determined empirically after several trials on a generated dataset), so default value is 50mm in order to have a security margin, even when dealing with co-registered files afterward.
    Returns
    -------
    bool :
        True if the volume is centered, False otherwise.

    Notes
    ------
    SPM has troubles to segment files if the center of the volume is not close from the origin of the world coordinate
    system. A series of experiment have been conducted: we take a volume whose center is on the origin of the world
    coordinate system. We add an offset using coordinates of affine matrix [0, 3], [1, 3], [2, 3] (or by modifying the
    header['srow_x'][3], header['srow_y'][3], header['srow_z'][3], this is strictly equivalent).
    It has been determined that volumes were still segmented with SPM when the L2 distance between origin and center of
    the volume did not exceed 100 mm. Above this distance, either the volume is either not segmented (SPM error), or the
    produced segmentation is wrong (not the shape of a brain anymore)
    """

    import numpy as np

    center = get_world_coordinate_of_center(nii_volume)

    distance_from_origin = np.linalg.norm(center, ord=2)

    return distance_from_origin < threshold_l2


def get_world_coordinate_of_center(nii_volume: PathLike) -> ndarray:
    """Extract the world coordinates of the center of the image.

    Parameters
    ---------
    nii_volume : PathLike
        path to nii volume

    Returns
    -------
    np.ndarray :
        coordinates in the world space
    References
    ------
    https://brainder.org/2012/09/23/the-nifti-file-format/
    """

    from os.path import isfile

    import nibabel as nib
    import numpy as np
    from nibabel.filebasedimages import ImageFileError

    from clinica.utils.exceptions import ClinicaException
    from clinica.utils.stream import cprint

    if not isfile(nii_volume):
        raise ClinicaException(
            f"The input {nii_volume} does not appear to be a path to a file."
        )

    try:
        orig_nifti = nib.load(nii_volume)
    except ImageFileError:
        cprint(
            msg=f"File {nii_volume} could not be read by nibabel. Is it a valid NIfTI file ?",
            lvl="warning",
        )
        return np.nan

    head = orig_nifti.header

    if isinstance(head, nib.freesurfer.mghformat.MGHHeader):
        # If MGH volume
        center_coordinates_world = vox_to_world_space_method_3_bis(
            head["dims"][0:3] / 2, head
        )
    else:
        # Standard NIfTI volume
        center_coordinates = get_center_volume(head)

        if head["qform_code"] > 0:
            center_coordinates_world = vox_to_world_space_method_2(
                center_coordinates, head
            )
        elif head["sform_code"] > 0:
            center_coordinates_world = vox_to_world_space_method_3(
                center_coordinates, head
            )
        elif head["sform_code"] == 0:
            center_coordinates_world = vox_to_world_space_method_1(
                center_coordinates, head
            )
        else:
            center_coordinates_world = np.nan
    return center_coordinates_world


def get_center_volume(header: Nifti1Header) -> ndarray:
    """Get the voxel coordinates of the center of the data, using header information.
    Parameters
    ----------
    header: Nifti1Header
            Contains image metadata
    Returns
    -------
    ndarray
        Voxel coordinates of the center of the volume
    """
    import numpy as np

    return np.array([x / 2 for x in header["dim"][1:4]])


def vox_to_world_space_method_1(
    coordinates_vol: ndarray, header: Nifti1Header
) -> ndarray:
    """Convert coordinates to world space
    Parameters
    ----------
    coordinates_vol: ndarray
        Coordinate in the volume (raw data)
    header: Nifti1Header
        Contains image metadata
    Returns
    -------
    ndarray
        Coordinates in the world space
    Notes
    -----
    This method is for compatibility with analyze and is not supposed to be used as the main orientation method. But it
    is used if sform_code = 0. The world coordinates are determined simply by scaling by the voxel size by their
    dimension stored in pixdim.
    References
    ----------
    https://brainder.org/2012/09/23/the-nifti-file-format/
    """
    import numpy as np

    return np.array(coordinates_vol) * np.array(
        [header["pixdim"][1], header["pixdim"][2], header["pixdim"][3]]
    )


def vox_to_world_space_method_2(
    coordinates_vol: ndarray, header: Nifti1Header
) -> ndarray:
    """Convert coordinates to world space (method 2)
    Parameters
    ----------
    coordinates_vol : ndarray
        Coordinate in the volume (raw data)
    header : Nifti1Header
        Image header containing metadata
    Returns
    -------
    ndarray
        Coordinates in the world space
    Notes
    -----
    This method is used when short qform_code is larger than zero. To get the coordinates, we multiply a rotation
    matrix (r_mat) by coordinates_vol, then perform Hadamard with pixel dimension pixdim (like in method 1). Then we add
    an offset (qoffset_x, qoffset_y, qoffset_z)
    """
    import numpy as np

    def get_r_matrix(h):
        """Get rotation matrix.
        More information here: https://brainder.org/2012/09/23/the-nifti-file-format/
        Parameters
        ----------
            h: header
        Returns
        -------
            Rotation matrix
        """
        b = h["quatern_b"]
        c = h["quatern_c"]
        d = h["quatern_d"]
        a = np.sqrt(1 - (b**2) - (c**2) - (d**2))
        r = np.zeros((3, 3))
        r[0, 0] = (a**2) + (b**2) - (c**2) - (d**2)
        r[0, 1] = 2 * ((b * c) - (a * d))
        r[0, 2] = 2 * ((b * d) + (a * c))
        r[1, 0] = 2 * ((b * c) + (a * d))
        r[1, 1] = (a**2) + (c**2) - (b**2) - (d**2)
        r[1, 2] = 2 * ((c * d) - (a * b))
        r[2, 0] = 2 * ((b * d) - (a * c))
        r[2, 1] = 2 * ((b * d) - (a * c))
        r[2, 2] = (a**2) + (d**2) - (b**2) - (c**2)
        return r

    i, j, k = coordinates_vol
    if header["qform_code"] > 0:
        r_mat = get_r_matrix(header)
    else:
        # Should never be reached
        raise ValueError("qform_code must be greater than 0 to use this method")
    q = header["pixdim"][0]
    if q not in [-1, 1]:
        print("q was " + str(q), ", now is 1")
        q = 1
    return np.dot(r_mat, np.array([i, j, q * k])) * np.array(
        header["pixdim"][1:4]
    ) + np.array([header["qoffset_x"], header["qoffset_y"], header["qoffset_z"]])


def vox_to_world_space_method_3(coordinates_vol: ndarray, header: Nifti1Header):
    """Convert coordinates to world space (method 3)
    Parameters
    ----------
    coordinates_vol : ndarray
        Coordinate in the volume (raw data)
    header : Nifti1Header
        Image header containing metadata
    Returns
    -------
    ndarray
        Coordinates in the world space
    Notes
    -----
    This method is used when sform_code is larger than zero. It relies on a full affine matrix, stored in the header in
    the fields srow_[x,y,y], to map voxel to world coordinates.
    When a nifti file is created with raw data and affine=..., this is this method that is used to decipher the
    voxel-to-world correspondence.
    """
    import numpy as np

    def get_aff_matrix(h: Nifti1Header) -> ndarray:
        """Get affine transformation matrix.
        Parameters
        ----------
            h : Nifti1Header
        Returns
        -------
        ndarray
            Affine transformation matrix
        References
        ----------
        https://brainder.org/2012/09/23/the-nifti-file-format/
        """
        mat = np.zeros((4, 4))
        mat[0, 0] = h["srow_x"][0]
        mat[0, 1] = h["srow_x"][1]
        mat[0, 2] = h["srow_x"][2]
        mat[0, 3] = h["srow_x"][3]
        mat[1, 0] = h["srow_y"][0]
        mat[1, 1] = h["srow_y"][1]
        mat[1, 2] = h["srow_y"][2]
        mat[1, 3] = h["srow_y"][3]
        mat[2, 0] = h["srow_z"][0]
        mat[2, 1] = h["srow_z"][1]
        mat[2, 2] = h["srow_z"][2]
        mat[2, 3] = h["srow_z"][3]
        mat[3, 3] = 1
        return mat

    if header["sform_code"] > 0:
        aff = get_aff_matrix(header)
    else:
        # Should never be reached
        raise ValueError("sform_code has a value > 0, so method 3 cannot be used")

    homogeneous_coord = np.concatenate(
        (np.array(coordinates_vol), np.array([1])), axis=0
    )
    return np.dot(aff, homogeneous_coord)[0:3]


def vox_to_world_space_method_3_bis(coordinates_vol: ndarray, header: Nifti1Header):
    """
    Convert coordinates to world space (method 3 bis).

    Parameters
    ----------
    coordinates_vol : ndarray
        Coordinate in the volume (raw data)
    header : Nifti1Header
        Image header containing metadata
    Returns
    -------
    ndarray
        Coordinates in the world space
    Notes
    -----
    This method relies on the same technique as method 3, but for images created by FreeSurfer (MGHImage, MGHHeader).
    """
    import numpy as np

    affine_transformation_matrix = header.get_affine()
    homogeneous_coord = np.concatenate(
        (np.array(coordinates_vol), np.array([1])), axis=0
    )
    return np.dot(affine_transformation_matrix, homogeneous_coord)[0:3]



import clinica.pipelines.engine as cpe


class T1VolumeCreateDartel(cpe.Pipeline):
    """T1VolumeCreateDartel - Create new Dartel template.

    Returns:
        A clinica pipeline object containing the T1VolumeCreateDartel pipeline.
    """

    def check_pipeline_parameters(self):
        """Check pipeline parameters."""
        from clinica.utils.group import check_group_label

        if "group_label" not in self.parameters.keys():
            raise KeyError("Missing compulsory group_label key in pipeline parameter.")
        self.parameters.setdefault("dartel_tissues", [1, 2, 3])

        check_group_label(self.parameters["group_label"])

    def check_custom_dependencies(self):
        """Check dependencies that can not be listed in the `info.json` file."""

    def get_input_fields(self):
        """Specify the list of possible inputs of this pipeline.

        Returns:
            A list of (string) input fields name.
        """
        return ["dartel_inputs"]

    def get_output_fields(self):
        """Specify the list of possible outputs of this pipeline.

        Returns:
            A list of (string) output fields name.
        """
        return ["final_template_file", "template_files", "dartel_flow_fields"]

    def build_input_node(self):
        """Build and connect an input node to the pipeline."""
        import os
        import sys

        import nipype.interfaces.utility as nutil
        import nipype.pipeline.engine as npe

        from clinica.utils.exceptions import ClinicaException
        from clinica.utils.input_files import t1_volume_dartel_input_tissue
        from clinica.utils.inputs import clinica_file_reader
        from clinica.utils.stream import cprint
        from clinica.utils.ux import (
            print_begin_image,
            print_groups_in_caps_directory,
            print_images_to_process,
        )

        representative_output = os.path.join(
            self.caps_directory,
            "groups",
            f"group-{self.parameters['group_label']}",
            "t1",
            f"group-{self.parameters['group_label']}_template.nii.gz",
        )
        if os.path.exists(representative_output):
            cprint(
                msg=(
                    f"DARTEL template for {self.parameters['group_label']} already exists. "
                    "Currently, Clinica does not propose to overwrite outputs for this pipeline."
                ),
                lvl="warning",
            )
            print_groups_in_caps_directory(self.caps_directory)
            sys.exit(0)

        # Check that there is at least 2 subjects
        if len(self.subjects) <= 1:
            raise ClinicaException(
                "This pipeline needs at least 2 images to create DARTEL "
                f"template but Clinica only found {len(self.subjects)}."
            )

        read_parameters_node = npe.Node(
            name="LoadingCLIArguments",
            interface=nutil.IdentityInterface(
                fields=self.get_input_fields(), mandatory_inputs=True
            ),
        )
        all_errors = []
        d_input = []
        for tissue_number in self.parameters["dartel_tissues"]:
            try:
                current_file, _ = clinica_file_reader(
                    self.subjects,
                    self.sessions,
                    self.caps_directory,
                    t1_volume_dartel_input_tissue(tissue_number),
                )
                d_input.append(current_file)
            except ClinicaException as e:
                all_errors.append(e)

        # Raise all errors if some happened
        if len(all_errors) > 0:
            error_message = "Clinica faced errors while trying to read files in your BIDS or CAPS directories.\n"
            for msg in all_errors:
                error_message += str(msg)
            raise RuntimeError(error_message)

        # d_input is a list of size len(self.parameters['dartel_tissues'])
        #     Each element of this list is a list of size len(self.subjects)
        read_parameters_node.inputs.dartel_inputs = d_input

        if len(self.subjects):
            print_images_to_process(self.subjects, self.sessions)
            cprint(
                "Computational time for DARTEL creation will depend on the number of images."
            )
            print_begin_image(f"group-{self.parameters['group_label']}")

        # fmt: off
        self.connect(
            [
                (read_parameters_node, self.input_node, [("dartel_inputs", "dartel_input_images")])
            ]
        )
        # fmt: on

    def build_output_node(self):
        """Build and connect an output node to the pipeline."""
        import os.path as op
        import re

        import nipype.interfaces.io as nio
        import nipype.pipeline.engine as npe

        from clinica.utils.filemanip import zip_nii

        # Writing flowfields into CAPS
        # ============================
        write_flowfields_node = npe.MapNode(
            name="write_flowfields_node",
            iterfield=["container", "flow_fields"],
            interface=nio.DataSink(infields=["flow_fields"]),
        )
        write_flowfields_node.inputs.base_directory = self.caps_directory
        write_flowfields_node.inputs.parameterization = False
        write_flowfields_node.inputs.container = [
            "subjects/"
            + self.subjects[i]
            + "/"
            + self.sessions[i]
            + "/t1/spm/dartel/group-"
            + self.parameters["group_label"]
            for i in range(len(self.subjects))
        ]
        write_flowfields_node.inputs.regexp_substitutions = [
            (r"(.*)_Template(\.nii(\.gz)?)$", r"\1\2"),
            (r"(.*)c1(sub-.*)(\.nii(\.gz)?)$", r"\1\2_segm-graymatter\3"),
            (r"(.*)c2(sub-.*)(\.nii(\.gz)?)$", r"\1\2_segm-whitematter\3"),
            (r"(.*)c3(sub-.*)(\.nii(\.gz)?)$", r"\1\2_segm-csf\3"),
            (r"(.*)c4(sub-.*)(\.nii(\.gz)?)$", r"\1\2_segm-bone\3"),
            (r"(.*)c5(sub-.*)(\.nii(\.gz)?)$", r"\1\2_segm-softtissue\3"),
            (r"(.*)c6(sub-.*)(\.nii(\.gz)?)$", r"\1\2_segm-background\3"),
            (r"(.*)r(sub-.*)(\.nii(\.gz)?)$", r"\1\2\3"),
            (r"(.*)_dartelinput(\.nii(\.gz)?)$", r"\1\2"),
            (
                r"(.*)flow_fields/u_(sub-.*)_segm-.*(\.nii(\.gz)?)$",
                r"\1\2_target-"
                + re.escape(self.parameters["group_label"])
                + r"_transformation-forward_deformation\3",
            ),
            (r"trait_added", r""),
        ]

        # Writing templates into CAPS
        # ===========================
        write_template_node = npe.Node(nio.DataSink(), name="write_template_node")
        write_template_node.inputs.parameterization = False
        write_template_node.inputs.base_directory = self.caps_directory
        write_template_node.inputs.container = op.join(
            "groups", f"group-{self.parameters['group_label']}", "t1"
        )
        write_template_node.inputs.regexp_substitutions = [
            (
                r"(.*)final_template_file/.*(\.nii(\.gz)?)$",
                r"\1group-"
                + re.escape(self.parameters["group_label"])
                + r"_template\2",
            ),
            (
                r"(.*)template_files/.*([0-9])(\.nii(\.gz)?)$",
                r"\1group-"
                + re.escape(self.parameters["group_label"])
                + r"_iteration-\2_template\3",
            ),
        ]

        # fmt: off
        self.connect(
            [
                (self.output_node, write_flowfields_node, [(("dartel_flow_fields", zip_nii, True), "flow_fields")]),
                (self.output_node, write_template_node, [(("final_template_file", zip_nii, True), "final_template_file"),
                                                         (("template_files", zip_nii, True), "template_files")],
                ),
            ]
        )
        # fmt: on

    def build_core_nodes(self):
        """Build and connect the core nodes of the pipeline."""
        import nipype.interfaces.spm as spm
        import nipype.interfaces.utility as nutil
        import nipype.pipeline.engine as npe

        from clinica.utils.filemanip import unzip_nii
        from clinica.utils.spm import spm_standalone_is_available, use_spm_standalone

        if spm_standalone_is_available():
            use_spm_standalone()

        # Unzipping
        # =========
        unzip_node = npe.MapNode(
            nutil.Function(
                input_names=["in_file"], output_names=["out_file"], function=unzip_nii
            ),
            name="unzip_node",
            iterfield=["in_file"],
        )

        # DARTEL template
        # ===============
        dartel_template = npe.Node(spm.DARTEL(), name="dartel_template")

        # Connection
        # ==========
        # fmt: off
        self.connect(
            [
                (self.input_node, unzip_node, [("dartel_input_images", "in_file")]),
                (unzip_node, dartel_template, [("out_file", "image_files")]),
                (dartel_template, self.output_node, [("dartel_flow_fields", "dartel_flow_fields"),
                                                     ("final_template_file", "final_template_file"),
                                                     ("template_files", "template_files")]),
            ]
        )
        # fmt: on



import abc


class Converter:
    __metaclass__ = abc.ABCMeta

    @abc.abstractmethod
    def convert_images(self, src, dst):
        pass

    @abc.abstractmethod
    def convert_clinical_data(self, src, dst):
        pass



"""Convert the NIFD dataset into BIDS."""

from os import PathLike
from typing import List


def convert_images(
    path_to_dataset: PathLike,
    bids_dir: PathLike,
    path_to_clinical: PathLike,
) -> List[PathLike]:
    """Convert the entire dataset in BIDS.

    Scans available files in the path_to_dataset,
    identifies the patients that have images described by the JSON file,
    converts the image with the highest quality for each category.
    """
    import clinica.iotools.bids_utils as bids

    from .oasis3_utils import (
        dataset_to_bids,
        intersect_data,
        read_clinical_data,
        read_imaging_data,
        write_bids,
    )

    # read the clinical data files
    dict_df = read_clinical_data(path_to_clinical)

    # makes a df of the imaging data
    imaging_data = read_imaging_data(path_to_dataset)

    # intersect the data
    imaging_data, df_small = intersect_data(imaging_data, dict_df)

    # build the tsv
    participants, sessions, scans = dataset_to_bids(imaging_data, df_small)

    written = write_bids(
        to=bids_dir,
        participants=participants,
        sessions=sessions,
        scans=scans,
        dataset_directory=path_to_dataset,
    )
    readme_data = {
        "link": "https://www.oasis-brains.org/#access",
        "desc": (
            "OASIS-3 is a retrospective compilation of data for 1378 participants that were collected across several "
            "ongoing projects through the WUSTL Knight ADRC over the course of 30years. Participants include 755 "
            "cognitively normal adults and 622 individuals at various stages of cognitive decline ranging in age from "
            "42-95yrs. All participants were assigned a new random identifier and all dates were removed and "
            "normalized to reflect days from entry into study. The dataset contains 2842 MR sessions which include "
            "T1w, T2w, FLAIR, ASL, SWI, time of flight, resting-state BOLD, and DTI sequences. Many of the MR sessions "
            "are accompanied by volumetric segmentation files produced through FreeSurfer processing. PET imaging from "
            "different tracers, PIB, AV45, and FDG, totaling over 2157 raw imaging scans and the accompanying "
            "post-processed files from the Pet Unified Pipeline (PUP) are also available in OASIS-3."
        ),
    }
    bids.write_modality_agnostic_files(
        study_name="OASIS-3", readme_data=readme_data, bids_dir=bids_dir
    )

    return written



"""This module contains FreeSurfer utilities."""


def generate_regional_measures_alt(
    segmentation_path, subject_id, atlas, output_dir=None
):
    """
    Read stats files located in
    <segmentation_path>/<subject_id>/stats/*.stats
    and generate TSV files in <segmentation_path>/regional_measures
    folder.

    Note: the .stats files contain both 1) a table with statistical
    information (e.g., structure volume) and 2) 'secondary' statistical
    information with all lines starting with the sentence '# Measure'.
    The .tsv files return the relevant statistical information from both
    sources.

    Args:
        segmentation_path (string): Path to the FreeSurfer segmentation.
        subject_id (string): Subject ID in the form sub-CLNC01_ses-M00, sub-CLNC01_long-M00M18 or
            sub-CLNC01_ses-M00.long.sub-CLNC01_long-M00M18
        output_dir (string): folder where the .tsv stats files will be
            stored. Will be [path_segmentation]/regional_measures if no
            dir is provided by the user
    """
    import errno
    import os

    import pandas

    from clinica.utils.freesurfer import (
        extract_image_id_from_longitudinal_segmentation,
        get_secondary_stats,
        write_tsv_file,
    )

    if segmentation_path != "":
        image_id = extract_image_id_from_longitudinal_segmentation(subject_id)
        prefix = image_id.participant_id
        if image_id.session_id:
            prefix = prefix + "_" + image_id.session_id
        if image_id.long_id:
            prefix = prefix + "_" + image_id.long_id

        stats_folder = os.path.join(
            os.path.expanduser(segmentation_path), subject_id, "stats"
        )
        if not os.path.isdir(stats_folder):
            raise IOError(
                "Image %s does not contain FreeSurfer segmentation"
                % prefix.replace("_", " | ")
            )

        if not output_dir:
            output_dir = os.path.join(segmentation_path, "regional_measures")
        try:
            os.makedirs(output_dir)
        except OSError as exception:
            # if dest_dir exists, go on, if its other error, raise
            if exception.errno != errno.EEXIST:
                raise

        # Generate TSV files for parcellation files
        #
        # Columns in ?h.BA.stats, ?h.aparc.stats or ?h.aparc.a2009s.stats file
        columns_parcellation = [
            "StructName",
            "NumVert",
            "SurfArea",
            "GrayVol",
            "ThickAvg",
            "ThickStd",
            "MeanCurv",
            "GausCurv",
            "FoldInd",
            "CurvInd",
        ]
        hemi_dict = {"left": "lh", "right": "rh"}
        info_dict = {
            "volume": "GrayVol",
            "thickness": "ThickAvg",
            "area": "SurfArea",
            "meancurv": "MeanCurv",
        }
        for atlas in [atlas]:
            stats_filename_dict = dict()
            df_dict = dict()
            # read both left and right .stats files
            for hemi in ("left", "right"):
                stats_filename_dict[hemi] = os.path.join(
                    stats_folder,
                    "{0}.{1}.stats".format(hemi_dict[hemi], atlas),
                )
                df_dict[hemi] = pandas.read_csv(
                    stats_filename_dict[hemi],
                    names=columns_parcellation,
                    comment="#",
                    header=None,
                    delimiter="\s+",
                    dtype=str,
                )
            # generate .tsv from 1) the table in .stats file and 2) the
            # secondary (commented out) information common to both 'left'
            # and 'right' .stats file
            for info in ("volume", "thickness", "area", "meancurv"):
                # Secondary information (common to 'left' and 'right')
                secondary_stats_dict = get_secondary_stats(
                    stats_filename_dict["left"], info
                )
                # Join primary and secondary information
                key_list = (
                    list("lh_" + df_dict["left"]["StructName"])
                    + list("rh_" + df_dict["right"]["StructName"])
                    + list(secondary_stats_dict.keys())
                )
                col_name = info_dict[info]
                value_list = (
                    list(df_dict["left"][col_name])
                    + list(df_dict["right"][col_name])
                    + list(secondary_stats_dict.values())
                )
                # Write .tsv
                write_tsv_file(
                    os.path.join(
                        output_dir,
                        "{0}_parcellation-{1}_{2}.tsv".format(prefix, atlas, info),
                    ),
                    key_list,
                    info,
                    value_list,
                )

        # Generate TSV files for segmentation files
        #
        # Columns in aseg.stats or wmparc.stats file
        columns_segmentation = [
            "Index",
            "SegId",
            "NVoxels",
            "Volume_mm3",
            "StructName",
            "normMean",
            "normStdDev",
            "normMin",
            "normMax",
            "normRange",
        ]

        # Parsing aseg.stats
        stats_filename = os.path.join(stats_folder, "aseg.stats")
        df = pandas.read_csv(
            stats_filename,
            comment="#",
            header=None,
            delimiter="\s+",
            dtype=str,
            names=columns_segmentation,
        )
        secondary_stats_dict = get_secondary_stats(stats_filename, "volume")
        key_list = list(df["StructName"]) + list(secondary_stats_dict.keys())
        value_list = list(df["Volume_mm3"]) + list(secondary_stats_dict.values())
        write_tsv_file(
            os.path.join(output_dir, prefix + "_segmentationVolumes.tsv"),
            key_list,
            "volume",
            value_list,
        )

        # Parsing wmparc.stats
        stats_filename = os.path.join(stats_folder, "wmparc.stats")
        df = pandas.read_csv(
            stats_filename,
            comment="#",
            header=None,
            delimiter="\s+",
            dtype=str,
            names=columns_segmentation,
        )
        secondary_stats_dict = get_secondary_stats(stats_filename, "volume")
        key_list = list(df["StructName"]) + list(secondary_stats_dict.keys())
        value_list = list(df["Volume_mm3"]) + list(secondary_stats_dict.values())
        write_tsv_file(
            os.path.join(output_dir, prefix + "_parcellation-wm_volume.tsv"),
            key_list,
            "volume",
            value_list,
        )


def extract_image_id_from_longitudinal_segmentation(freesurfer_id):
    """Extract image ID from longitudinal segmentation folder.

    This function will extract participant, session and longitudinal ID from `freesurfer_id`.

    Example:
        >>> from clinica.utils.freesurfer import extract_image_id_from_longitudinal_segmentation
        >>> extract_image_id_from_longitudinal_segmentation('sub-CLNC01_ses-M00')
        image_id(participant_id='sub-CLNC01', session_id='ses-M00', long_id='')
        >>> extract_image_id_from_longitudinal_segmentation('sub-CLNC01_long-M0018')
        image_id(participant_id='sub-CLNC01', session_id='', long_id='long-M0018')
        >>> extract_image_id_from_longitudinal_segmentation('sub-CLNC01_ses-M00.long.sub-CLNC01_long-M00M18')
        image_id(participant_id='sub-CLNC01', session_id='ses-M00', long_id='long-M00M18')
    """
    from collections import namedtuple

    image_id = namedtuple("image_id", ["participant_id", "session_id", "long_id"])

    # Case 'sub-CLNC01_ses-M00.long.sub-CLNC01_long-M00M18'
    if ".long." in freesurfer_id:
        participant_id = freesurfer_id.split(".long.")[0].split("_")[0]
        session_id = freesurfer_id.split(".long.")[0].split("_")[1]
        long_id = freesurfer_id.split(".long.")[1].split("_")[1]
    # Case 'sub-CLNC01_long-M00M18'
    elif "long-" in freesurfer_id:
        participant_id = freesurfer_id.split("_")[0]
        session_id = ""
        long_id = freesurfer_id.split("_")[1]
    # Case 'sub-CLNC01_ses-M00'
    else:
        participant_id = freesurfer_id.split("_")[0]
        session_id = freesurfer_id.split("_")[1]
        long_id = ""

    return image_id(participant_id, session_id, long_id)


def get_secondary_stats(stats_filename, info_type):
    """Read the 'secondary' statistical info from .stats file.

    Extract the information from .stats file that is commented out
    (lines starting with '# Measure' prefix) and does not appear in the
    table at the end of the document.

    Args:
        stats_filename (string): path to the .stats file
        info_type (string): 'volume', 'thickness' or 'area'

    Returns:
        secondary_stats_dict (dictionary of float): keys are regions of
            the brain, associated values are the corresponding
            volume/thickness/area depending on the input info type
    """
    # initialise structure containing the secondary statistical info
    secondary_stats_dict = dict()

    # currently no additional information is provided by .stats file for
    # the mean curvature
    if info_type != "meancurv":
        # define how lines are supposed to end in the stats file, depending
        # on the type of information that is searched for
        endline_dict = dict()
        endline_dict["volume"] = "mm^3"
        endline_dict["thickness"] = "mm"
        endline_dict["area"] = "mm^2"

        # define keywords that are supposed to appear in commented lines
        # containing statistical information
        info_keyword_dict = dict()
        info_keyword_dict["volume"] = ["volume", "Volume"]
        info_keyword_dict["area"] = ["area", "Area"]
        info_keyword_dict["thickness"] = ["thickness", "Thickness"]

        # read stats file line by line and only keep relevant lines
        with open(stats_filename, "r") as stats_file:
            stats = stats_file.read()
        stats_line_list = stats.splitlines()
        for stats_line in stats_line_list:
            startswith_condition = stats_line.startswith("# Measure")
            endswith_condition = stats_line.endswith(endline_dict[info_type])
            if (startswith_condition) and (endswith_condition):
                stats_line_word_list = stats_line.replace(",", "").split()
                # sanity check: make sure any sensible variation of
                # 'volume', 'thickness' or 'area' appears inside the line
                if any(x in stats_line_word_list for x in info_keyword_dict[info_type]):
                    # add info
                    info_region = stats_line_word_list[2]
                    info_value = stats_line_word_list[-2]
                    secondary_stats_dict[info_region] = info_value

    return secondary_stats_dict


def generate_regional_measures(segmentation_path, subject_id, output_dir=None):
    """General regional measures files from FreeSurfer segmentation.

    Read stats files located in <segmentation_path>/<subject_id>/stats/*.stats
    and generate TSV files in <segmentation_path>/regional_measures folder.

    Note: the .stats files contain both 1) a table with statistical
    information (e.g., structure volume) and 2) 'secondary' statistical
    information with all lines starting with the sentence '# Measure'.
    The .tsv files return the relevant statistical information from both
    sources.

    Args:
        segmentation_path (string): Path to the FreeSurfer segmentation.
        subject_id (string): Subject ID in the form sub-CLNC01_ses-M00, sub-CLNC01_long-M00M18 or
            sub-CLNC01_ses-M00.long.sub-CLNC01_long-M00M18
        output_dir (string): folder where the .tsv stats files will be
            stored. Will be [path_segmentation]/regional_measures if no
            dir is provided by the user
    """
    import os

    import pandas

    from clinica.utils.freesurfer import write_tsv_file

    image_id = extract_image_id_from_longitudinal_segmentation(subject_id)
    prefix = image_id.participant_id
    if image_id.session_id:
        prefix = f"{prefix}_{image_id.session_id}"
    if image_id.long_id:
        prefix = f"{prefix}_{image_id.long_id}"

    stats_folder = os.path.join(
        os.path.expanduser(segmentation_path), subject_id, "stats"
    )

    if not os.path.isdir(stats_folder):
        raise IOError(
            f"Image {prefix.replace('_', ' | ')} does not contain FreeSurfer segmentation"
        )

    if not output_dir:
        output_dir = os.path.join(segmentation_path, "regional_measures")
    os.makedirs(output_dir, exist_ok=True)

    # Generate TSV files for parcellation files
    #
    # Columns in ?h.BA.stats, ?h.aparc.stats or ?h.aparc.a2009s.stats file
    columns_parcellation = [
        "StructName",
        "NumVert",
        "SurfArea",
        "GrayVol",
        "ThickAvg",
        "ThickStd",
        "MeanCurv",
        "GausCurv",
        "FoldInd",
        "CurvInd",
    ]
    hemi_dict = {"left": "lh", "right": "rh"}
    atlas_dict = {"desikan": "aparc", "destrieux": "aparc.a2009s", "ba": "BA_exvivo"}
    info_dict = {
        "volume": "GrayVol",
        "thickness": "ThickAvg",
        "area": "SurfArea",
        "meancurv": "MeanCurv",
    }
    for atlas in ("desikan", "destrieux", "ba"):
        stats_filename_dict = dict()
        df_dict = dict()
        # read both left and right .stats files
        for hemi in ("left", "right"):
            stats_filename_dict[hemi] = os.path.join(
                stats_folder, "{0}.{1}.stats".format(hemi_dict[hemi], atlas_dict[atlas])
            )
            df_dict[hemi] = pandas.read_csv(
                stats_filename_dict[hemi],
                names=columns_parcellation,
                comment="#",
                header=None,
                delimiter="\s+",
                dtype=str,
            )
        # generate .tsv from 1) the table in .stats file and 2) the
        # secondary (commented out) information common to both 'left'
        # and 'right' .stats file
        for info in ("volume", "thickness", "area", "meancurv"):
            # Secondary information (common to 'left' and 'right')
            secondary_stats_dict = get_secondary_stats(
                stats_filename_dict["left"], info
            )
            # Join primary and secondary information
            key_list = (
                list("lh_" + df_dict["left"]["StructName"])
                + list("rh_" + df_dict["right"]["StructName"])
                + list(secondary_stats_dict.keys())
            )
            col_name = info_dict[info]
            value_list = (
                list(df_dict["left"][col_name])
                + list(df_dict["right"][col_name])
                + list(secondary_stats_dict.values())
            )
            # Write .tsv
            write_tsv_file(
                os.path.join(
                    output_dir,
                    f"{prefix}_parcellation-{atlas}_{info}.tsv",
                ),
                key_list,
                info,
                value_list,
            )

    # Generate TSV files for segmentation files
    #
    # Columns in aseg.stats or wmparc.stats file
    columns_segmentation = [
        "Index",
        "SegId",
        "NVoxels",
        "Volume_mm3",
        "StructName",
        "normMean",
        "normStdDev",
        "normMin",
        "normMax",
        "normRange",
    ]

    # Parsing aseg.stats
    stats_filename = os.path.join(stats_folder, "aseg.stats")
    df = pandas.read_csv(
        stats_filename,
        comment="#",
        header=None,
        delimiter="\s+",
        dtype=str,
        names=columns_segmentation,
    )
    secondary_stats_dict = get_secondary_stats(stats_filename, "volume")
    key_list = list(df["StructName"]) + list(secondary_stats_dict.keys())
    value_list = list(df["Volume_mm3"]) + list(secondary_stats_dict.values())
    write_tsv_file(
        os.path.join(output_dir, prefix + "_segmentationVolumes.tsv"),
        key_list,
        "volume",
        value_list,
    )

    # Parsing wmparc.stats
    stats_filename = os.path.join(stats_folder, "wmparc.stats")
    df = pandas.read_csv(
        stats_filename,
        comment="#",
        header=None,
        delimiter="\s+",
        dtype=str,
        names=columns_segmentation,
    )
    secondary_stats_dict = get_secondary_stats(stats_filename, "volume")
    key_list = list(df["StructName"]) + list(secondary_stats_dict.keys())
    value_list = list(df["Volume_mm3"]) + list(secondary_stats_dict.values())
    write_tsv_file(
        os.path.join(output_dir, prefix + "_parcellation-wm_volume.tsv"),
        key_list,
        "volume",
        value_list,
    )


def write_tsv_file(out_filename, name_list, scalar_name, scalar_list):
    """Write a .tsv file with list of keys and values.

    Args:
        out_filename (string): name of the .tsv file
        name_list (list of string): list of keys
        scalar_name (string): 'volume', 'thickness', 'area' or
            'meanCurv'. Not used for now. Might be used as part of a
            pandas data frame
        scalar_list (list of float): list of values corresponding to the keys
    """
    import warnings

    import pandas

    try:
        data = pandas.DataFrame({"label_name": name_list, "label_value": scalar_list})
        data.to_csv(out_filename, sep="\t", index=False, encoding="utf-8")
    except Exception as exception:
        warnings.warn("Impossible to save {0} file".format(out_filename))
        raise exception

    return out_filename


def check_flags(in_t1w, recon_all_args):
    """Check `recon_all_args` flags for `in_t1w` image.

    Currently, this function only adds '-cw256' if the FOV of `in_t1w` is greater than 256.
    """
    import nibabel as nib

    # from clinica.utils.stream import cprint

    f = nib.load(in_t1w)
    voxel_size = f.header.get_zooms()
    t1_size = f.header.get_data_shape()
    if (
        (voxel_size[0] * t1_size[0] > 256)
        or (voxel_size[1] * t1_size[1] > 256)
        or (voxel_size[2] * t1_size[2] > 256)
    ):
        # cprint(f"Setting MRI Convert to crop images to 256 FOV for {in_t1w} file.")
        optional_flag = " -cw256"
    else:
        # cprint(f"No need to add -cw256 flag for {in_t1w} file.")
        optional_flag = ""
    flags = "{0}".format(recon_all_args) + optional_flag

    return flags



"""The 'clinica' executable command line, installed with the clinica package, calls this module.

The aim of this module is to execute pipelines from the command line,
and gives to the user some other utils to work with the pipelines.
"""
import os
import sys

import click

import clinica
from clinica.engine.template import cli as generate_cli
from clinica.iotools.converters.cli import cli as convert_cli
from clinica.iotools.utils.cli import cli as iotools_cli
from clinica.pipelines.cli import cli as run_cli
from clinica.utils.exceptions import ClinicaException
from clinica.utils.stream import cprint

CONTEXT_SETTINGS = dict(
    # Extend content width to avoid shortening of pipeline help.
    max_content_width=160,
    # Display help string with -h, in addition to --help.
    help_option_names=["-h", "--help"],
)


def setup_logging(verbose: bool = False) -> None:
    """Setup Clinica's logging facilities.

    Args:
        verbosity (int): The desired level of verbosity for logging.
            (0 (default): WARNING, 1: INFO, 2: DEBUG)
    """
    import logging
    import sys
    from logging.handlers import RotatingFileHandler as RFHandler

    import nipype
    from colorlog import ColoredFormatter, StreamHandler

    logging_level = "DEBUG" if verbose else "INFO"

    # Root logger configuration.
    root_logger = logging.getLogger()
    root_logger.setLevel(logging_level)

    # Clinica logger configuration.
    clinica_logger = logging.getLogger("clinica")
    clinica_logger.setLevel(logging_level)
    console_handler = StreamHandler(stream=sys.stdout)
    console_handler.setFormatter(
        ColoredFormatter("%(log_color)s%(asctime)s:%(levelname)s:%(message)s")
    )
    clinica_logger.addHandler(console_handler)

    # Nipype logger configuration.
    # Monkey-patch nipype to use Python's RFH logger.
    nipype.utils.logger.RFHandler = RFHandler
    # Setup debug logging to file.
    nipype.config.enable_debug_mode()
    nipype.config.update_config(
        {
            "logging": {
                "log_directory": os.getenv("CLINICA_LOGGING_DIR", os.getcwd()),
                "log_to_file": True,
            },
            "execution": {"check_version": False},
        },
    )
    nipype.logging.update_logging(nipype.config)
    # Disable nipype logging to console.
    nipype_logger = logging.getLogger("nipype")
    nipype_logger.removeHandler(nipype_logger.handlers[0])
    nipype_logger.addHandler(logging.NullHandler())


@click.group(context_settings=CONTEXT_SETTINGS)
@click.version_option(version=clinica.__version__)
@click.option("-v", "--verbose", is_flag=True, help="Increase logging verbosity.")
def cli(verbose: bool) -> None:
    setup_logging(verbose=verbose)


cli.add_command(convert_cli)
cli.add_command(generate_cli)
cli.add_command(iotools_cli)
cli.add_command(run_cli)


def main() -> None:
    try:
        cli()
    except ClinicaException as e:
        cprint(msg=e, lvl="error")
        sys.exit(1)


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['clinica',
 'clinica.engine',
 'clinica.iotools',
 'clinica.iotools.converters',
 'clinica.iotools.converters.adni_to_bids',
 'clinica.iotools.converters.adni_to_bids.adni_modalities',
 'clinica.iotools.converters.aibl_to_bids',
 'clinica.iotools.converters.habs_to_bids',
 'clinica.iotools.converters.nifd_to_bids',
 'clinica.iotools.converters.oasis3_to_bids',
 'clinica.iotools.converters.oasis_to_bids',
 'clinica.iotools.converters.ukb_to_bids',
 'clinica.iotools.utils',
 'clinica.lib',
 'clinica.pipelines',
 'clinica.pipelines.cli_param',
 'clinica.pipelines.deeplearning_prepare_data',
 'clinica.pipelines.dwi_connectome',
 'clinica.pipelines.dwi_dti',
 'clinica.pipelines.dwi_preprocessing_using_fmap',
 'clinica.pipelines.dwi_preprocessing_using_t1',
 'clinica.pipelines.machine_learning',
 'clinica.pipelines.machine_learning_spatial_svm',
 'clinica.pipelines.pet_linear',
 'clinica.pipelines.pet_surface',
 'clinica.pipelines.pet_volume',
 'clinica.pipelines.statistics_surface',
 'clinica.pipelines.statistics_volume',
 'clinica.pipelines.statistics_volume_correction',
 'clinica.pipelines.t1_freesurfer',
 'clinica.pipelines.t1_freesurfer_atlas',
 'clinica.pipelines.t1_freesurfer_longitudinal',
 'clinica.pipelines.t1_linear',
 'clinica.pipelines.t1_volume',
 'clinica.pipelines.t1_volume_create_dartel',
 'clinica.pipelines.t1_volume_dartel2mni',
 'clinica.pipelines.t1_volume_existing_template',
 'clinica.pipelines.t1_volume_parcellation',
 'clinica.pipelines.t1_volume_register_dartel',
 'clinica.pipelines.t1_volume_tissue_segmentation',
 'clinica.pydra',
 'clinica.pydra.pet_volume',
 'clinica.pydra.t1_linear',
 'clinica.pydra.t1_volume',
 'clinica.pydra.t1_volume.create_dartel',
 'clinica.pydra.t1_volume.tissue_segmentation',
 'clinica.utils']

package_data = \
{'': ['*'],
 'clinica': ['resources/*',
             'resources/atlases/*',
             'resources/fmri/*',
             'resources/mappings/*',
             'resources/masks/*',
             'resources/templates/pipeline_template/*'],
 'clinica.iotools': ['data/*'],
 'clinica.lib': ['clinicasurfstat/*',
                 'clinicasurfstat/SurfStat/*',
                 'clinicasurfstat/SurfStat/@random/*',
                 'clinicasurfstat/SurfStat/@term/*']}

install_requires = \
['argcomplete>=1.9.4,<2.0.0',
 'attrs>=20.1.0',
 'brainstat>=0.3.6,<0.4.0',
 'cattrs>=1.9.0,<2.0.0',
 'click-option-group>=0.5,<0.6',
 'click>=8,<9',
 'colorlog>=5,<6',
 'fsspec',
 'jinja2>=3,<4',
 'joblib>=1.2.0,<2.0.0',
 'matplotlib',
 'networkx',
 'nibabel>=3.2.2,<4.0.0',
 'niflow-nipype1-workflows',
 'nilearn[plotting]>=0.9.2,<0.10.0',
 'nipype>=1.7.1,<1.8.3',
 'numpy>=1.17,<2.0',
 'openpyxl',
 'pandas>=1.4,<2.0',
 'pybids>=0.15.1,<0.16.0',
 'pydicom',
 'pydra-bids>=0.0.3,<0.0.4',
 'pydra-nipype1',
 'pydra>=0.20,<0.21',
 'scikit-image>=0.19,<0.20',
 'scikit-learn>=1.0,<2.0',
 'scipy>=1.7,<2.0',
 'xgboost',
 'xlrd',
 'xvfbwrapper']

entry_points = \
{'console_scripts': ['clinica = clinica.cmdline:main']}

setup_kwargs = {
    'name': 'clinica',
    'version': '0.7.3',
    'description': 'Software platform for clinical neuroimaging studies',
    'long_description': '<!--(http://www.clinica.run/img/clinica_brainweb.png)-->\n<!-- markdownlint-disable MD033 -->\n\n<h1 align="center">\n  <a href="http://www.clinica.run">\n    <img src="http://www.clinica.run/assets/images/clinica-icon-257x257.png" alt="Logo" width="120" height="120">\n  </a>\n  <br/>\n  Clinica\n</h1>\n\n<p align="center"><strong>Software platform for clinical neuroimaging studies</strong></p>\n\n<p align="center">\n  <a href="https://ci.inria.fr/clinica-aramis/job/clinica/job/dev/">\n    <img src="https://ci.inria.fr/clinica-aramis/buildStatus/icon?job=clinica%2Fdev" alt="Build Status">\n  </a>\n  <a href="https://badge.fury.io/py/clinica">\n    <img src="https://badge.fury.io/py/clinica.svg" alt="PyPI version">\n  </a>\n  <a href="https://pypi.org/project/clinica">\n    <img src="https://img.shields.io/pypi/pyversions/clinica" alt="Supported Python versions">\n  </a>\n  <a href="https://aramislab.paris.inria.fr/clinica/docs/public/latest/Installation/">\n  </a>\n  <a href="https://aramislab.paris.inria.fr/clinica/docs/public/latest/Installation/">\n    <img src="https://anaconda.org/aramislab/clinica/badges/platforms.svg" alt="platform">\n  </a>\n  <a href="https://github.com/psf/black">\n    <img src="https://img.shields.io/badge/code%20style-black-000000.svg" alt="Code style: black">\n  </a>\n</p>\n\n<p align="center">\n  <a href="http://www.clinica.run">Homepage</a> |\n  <a href="https://aramislab.paris.inria.fr/clinica/docs/public/latest/">Documentation</a> |\n  <a href="https://doi.org/10.3389/fninf.2021.689675">Paper</a> |\n  <a href="https://github.com/aramis-lab/clinica/discussions">Forum</a> |\n  See also:\n  <a href="#related-repositories">AD-ML</a>,\n  <a href="#related-repositories">AD-DL</a>,\n  <a href="#related-repositories">ClinicaDL</a>\n</p>\n\n## About The Project\n\nClinica is a software platform for clinical research studies involving patients\nwith neurological and psychiatric diseases and the acquisition of multimodal\ndata (neuroimaging, clinical and cognitive evaluations, genetics...),\nmost often with longitudinal follow-up.\n\nClinica is command-line driven and written in Python.\nIt uses the [Nipype](https://nipype.readthedocs.io/) system for pipelining and combines\nwidely-used software packages for neuroimaging data analysis\n([ANTs](http://stnava.github.io/ANTs/),\n[FreeSurfer](https://surfer.nmr.mgh.harvard.edu/),\n[FSL](https://fsl.fmrib.ox.ac.uk/fsl/fslwiki),\n[MRtrix](https://www.mrtrix.org/),\n[PETPVC](https://github.com/UCL/PETPVC),\n[SPM](https://www.fil.ion.ucl.ac.uk/spm/)), machine learning\n([Scikit-learn](https://scikit-learn.org/stable/)) and the [BIDS\nstandard](http://bids-specification.readthedocs.io/) for data organization.\n\nClinica provides tools to convert publicly available neuroimaging datasets into\nBIDS, namely:\n\n- [ADNI: Alzheimer’s Disease Neuroimaging Initiative](https://aramislab.paris.inria.fr/clinica/docs/public/latest/Converters/ADNI2BIDS/)\n- [AIBL: Australian Imaging, Biomarker & Lifestyle Flagship Study of Ageing](https://aramislab.paris.inria.fr/clinica/docs/public/latest/Converters/AIBL2BIDS/)\n- [HABS: Harvard Aging Brain Study](https://aramislab.paris.inria.fr/clinica/docs/public/latest/Converters/HABS2BIDS/)\n- [NIFD: Neuroimaging in Frontotemporal Dementia](https://aramislab.paris.inria.fr/clinica/docs/public/latest/Converters/NIFD2BIDS/)\n- [OASIS: Open Access Series of Imaging Studies](https://aramislab.paris.inria.fr/clinica/docs/public/latest/Converters/OASIS2BIDS/)\n- [OASIS-3: Longitudinal Neuroimaging, Clinical, and Cognitive Dataset for Normal Aging and Alzheimer’s Disease](https://aramislab.paris.inria.fr/clinica/docs/public/latest/Converters/OASIS3TOBIDS/)\n\nClinica can process any BIDS-compliant dataset with a set of complex processing\npipelines involving different software packages for the analysis of\nneuroimaging data (T1-weighted MRI, diffusion MRI and PET data).\nIt also provides integration between feature extraction and statistics, machine\nlearning or deep learning.\n\n![ClinicaPipelines](http://www.clinica.run/img/Clinica_Pipelines_A4_2021-04-02_75dpi.jpg)\n\nClinica is also showcased as a framework for the reproducible classification of\nAlzheimer\'s disease using\n[machine learning](https://github.com/aramis-lab/AD-ML) and\n[deep learning](https://github.com/aramis-lab/clinicadl).\n\n## Getting Started\n\n> Full instructions for installation and additional information can be found in\nthe [user documentation](https://aramislab.paris.inria.fr/clinica/docs/public/latest/).\n\nClinica currently supports macOS and Linux.\nIt can be installed by typing the following command:\n\n```sh\npip install clinica\n```\n\nTo avoid conflicts with other versions of the dependency packages installed by pip, it is strongly recommended to create a virtual environment before the installation.\nFor example, use [Conda](https://docs.conda.io/en/latest/miniconda.html), to create a virtual\nenvironment and activate it before installing clinica (you can also use\n`virtualenv`):\n\n```sh\nconda create --name clinicaEnv python=3.8\nconda activate clinicaEnv\n```\n\nDepending on the pipeline that you want to use, you need to install pipeline-specific interfaces.\nNot all the dependencies are necessary to run Clinica.\nPlease refer to this [page](https://aramislab.paris.inria.fr/clinica/docs/public/latest/Third-party/)\nto determine which third-party libraries you need to install.\n\n## Example\n\nDiagram illustrating the Clinica pipelines involved when performing a group\ncomparison of FDG PET data projected on the cortical surface between patients\nwith Alzheimer\'s disease and healthy controls from the ADNI database:\n\n![ClinicaExample](http://www.clinica.run/img/Clinica_Example_2021-04-02_75dpi.jpg)\n\n1. Clinical and neuroimaging data are downloaded from the ADNI website and data\n   are converted into BIDS with the [`adni-to-bids`\n   converter](https://aramislab.paris.inria.fr/clinica/docs/public/latest/Converters/ADNI2BIDS/).\n2. Estimation of the cortical and white surface is then produced by the\n   [`t1-freesurfer`\n   pipeline](https://aramislab.paris.inria.fr/clinica/docs/public/latest/Pipelines/T1_FreeSurfer/).\n3. FDG PET data can be projected on the subject’s cortical surface and\n   normalized to the FsAverage template from FreeSurfer using the\n   [`pet-surface` pipeline](https://aramislab.paris.inria.fr/clinica/docs/public/latest/Pipelines/PET_Surface/).\n4. TSV file with demographic information of the population studied is given to\n   the [`statistics-surface`\n   pipeline](https://aramislab.paris.inria.fr/clinica/docs/public/latest/Pipelines/Stats_Surface/) to generate\n   the results of the group comparison.\n\n> For more examples and details, please refer to the\n> [Documentation](https://aramislab.paris.inria.fr/clinica/docs/public/latest/).\n\n## Support\n\n- Check for [past answers](https://groups.google.com/forum/#!forum/clinica-user) in the old Clinica Google Group\n- Start a [discussion](https://github.com/aramis-lab/clinica/discussions) on Github\n- Report an [issue](https://github.com/aramis-lab/clinica/issues) on GitHub\n\n## Contributing\n\nWe encourage you to contribute to Clinica!\nPlease check out the [Contributing to Clinica guide](CONTRIBUTING.md) for\nguidelines about how to proceed.  Do not hesitate to ask questions if something\nis not clear for you, report an issue, etc.\n\n## License\n\nThis software is distributed under the MIT License.\nSee [license file](https://github.com/aramis-lab/clinica/blob/dev/LICENSE.txt)\nfor more information.\n\n## Citing us\n\n- Routier, A., Burgos, N., Díaz, M., Bacci, M., Bottani, S., El-Rifai O., Fontanella, S., Gori, P., Guillon, J., Guyot, A., Hassanaly, R., Jacquemont, T.,  Lu, P., Marcoux, A.,  Moreau, T., Samper-González, J., Teichmann, M., Thibeau-Sutre, E., Vaillant G., Wen, J., Wild, A., Habert, M.-O., Durrleman, S., and Colliot, O.:\n*Clinica: An Open Source Software Platform for Reproducible Clinical Neuroscience Studies* Frontiers in Neuroinformatics, 2021\n[doi:10.3389/fninf.2021.689675](https://doi.org/10.3389/fninf.2021.689675)\n\n## Related Repositories\n\n- [AD-DL: Classification of Alzheimer\'s disease status with convolutional neural networks](https://github.com/aramis-lab/AD-DL).\n- [AD-ML: Framework for the reproducible classification of Alzheimer\'s disease using\nmachine learning](https://github.com/aramis-lab/AD-ML).\n- [ClinicaDL: Framework for the reproducible processing of neuroimaging data with deep learning methods](https://github.com/aramis-lab/clinicadl).\n',
    'author': 'ARAMIS Lab',
    'author_email': 'None',
    'maintainer': 'Clinica developers',
    'maintainer_email': 'clinica-user@inria.fr',
    'url': 'https://www.clinica.run',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<3.11',
}


setup(**setup_kwargs)



"""This module contains utilities for PET data handling."""
import os
import typing as ty
from enum import Enum

import pandas as pd


class Tracer(str, Enum):
    """BIDS label for PET tracers.

    Follows the convention proposed in the PET section of the BIDS specification.

    See: https://bids-specification.readthedocs.io/en/stable/04-modality-specific-files/09-positron-emission-tomography.html
    """

    PIB = "11CPIB"
    AV1451 = "18FAV1451"
    AV45 = "18FAV45"
    FBB = "18FFBB"
    FDG = "18FFDG"
    FMM = "18FFMM"


def read_psf_information(
    pvc_psf_tsv: os.PathLike,
    subject_ids: ty.List[str],
    session_ids: ty.List[str],
    pet_tracer: str,
) -> ty.List[ty.List[int]]:
    """Read PSF information from TSV file.

    Parameters
    ----------
    pvc_psf_tsv : PathLike
        Path to the TSV file containing the following columns:
        'participant_id', 'session_id', 'acq_label', 'psf_x',
        'psf_y', and 'psf_z'

    subject_ids : List[str]
        List of participant IDs.
        ex: ['sub-CLNC01', 'sub-CLNC01']

        .. warning::
            Must have the same length as `session_ids`.

    session_ids : List[str]
        List of session IDs.
        ex: ['ses-M00', 'ses-M18']

        .. warning::
            Must have the same length as `subject_ids`.

    pet_tracer : str
        Tracer we want to select in the 'acq_label' column.
        Other tracers will not be read in this function

    Returns
    -------
    psf : List[List[int]]
        The PSF information following [subject_ids, session_ids] order.

    Examples
    --------
    Example of pvc_psf_tsv:

    participant_id    session_id     acq_label     psf_x    psf_y    psf_z
    sub-CLNC01        ses-M00        FDG           8        9        10
    sub-CLNC01        ses-M18        FDG           8        9        10
    sub-CLNC01        ses-M00        AV45          7        6        5
    sub-CLNC02        ses-M00        FDG           8        9        10
    sub-CLNC03        ses-M00        FDG           8        9        10
    """
    valid_columns = {
        "participant_id",
        "session_id",
        "acq_label",
        "psf_x",
        "psf_y",
        "psf_z",
    }
    psf_df = pd.read_csv(pvc_psf_tsv, sep="\t")
    diff = valid_columns.symmetric_difference(set(psf_df.columns))
    if len(diff) > 0:
        raise IOError(
            f"The file {pvc_psf_tsv} must contain the following columns (separated by tabulations):\n"
            f"participant_id, session_id, acq_label, psf_x, psf_y, psf_z\n"
            f"{str(list(psf_df.columns))}\n"
            f"Pay attention to the spaces (there should be none)."
        )
    psf = []
    for sub, ses in zip(subject_ids, session_ids):
        result = psf_df.query(
            f"participant_id == '{sub}' and session_id == '{ses}' and acq_label == '{pet_tracer}'"
        )
        if len(result) == 0:
            raise RuntimeError(
                f"Subject {sub} with session {ses} and tracer {pet_tracer} "
                f"that you want to proceed was not found in the TSV file containing "
                f"PSF specifications ({pvc_psf_tsv})."
            )
        if len(result) > 1:
            raise RuntimeError(
                f"Subject {sub} with session {ses} and tracer {pet_tracer} "
                f"that you want to proceed was found multiple times "
                f"in the TSV file containing PSF specifications ({pvc_psf_tsv})."
            )
        psf.append(result[["psf_x", "psf_y", "psf_z"]].values.tolist()[0])

    return psf


LIST_SUVR_REFERENCE_REGIONS = ["pons", "cerebellumPons", "pons2", "cerebellumPons2"]


def get_suvr_mask(suvr_reference_region: str) -> os.PathLike:
    """Returns the path to the SUVR mask from SUVR reference region label.

    Parameters
    ----------
    suvr_reference_region : str
        The label of the SUVR reference region.
        Supported labels are: 'pons', 'cerebellumPons', 'pons2', and 'cerebellumPons2'

    Returns
    -------
    PathLike :
        The path to the SUVR mask.
    """
    from pathlib import Path

    current_dir = Path(os.path.realpath(__file__))
    masks_dir = current_dir.parent.parent / "resources" / "masks"

    suvr_reference_region_labels_to_filenames = {
        "pons": "region-pons_eroded-6mm_mask.nii.gz",
        "cerebellumPons": "region-cerebellumPons_eroded-6mm_mask.nii.gz",
        "pons2": "region-pons_remove-extrabrain_eroded-2it_mask.nii.gz",
        "cerebellumPons2": "region-cerebellumPons_remove-extrabrain_eroded-3it_mask.nii.gz",
    }
    if suvr_reference_region not in suvr_reference_region_labels_to_filenames:
        raise ValueError(
            f"SUVR reference region label {suvr_reference_region} is not supported. "
            f"Supported values are : {list(suvr_reference_region_labels_to_filenames.keys())}."
        )
    return masks_dir / suvr_reference_region_labels_to_filenames[suvr_reference_region]



import typing as ty
from pathlib import PurePath

from pydra.mark import annotate, task


@task
@annotate({"return": {"suvr_pet_path": PurePath}})
def normalize_to_reference_task(pet_image: PurePath, region_mask: PurePath) -> PurePath:
    """Pydra task for normalizing the provided `pet_image`.

    .. note::
        Please refer to the documentation of function
        `clinica.pipelines.pet_volume.pet_volume_utils.normalize_to_reference`.
    """
    from clinica.pipelines.pet_volume.pet_volume_utils import normalize_to_reference

    return normalize_to_reference(pet_image, region_mask)


@task
@annotate({"return": {"out_mask": PurePath}})
def create_binary_mask_task(
    tissues: ty.List[PurePath],
    threshold: float = 0.3,
) -> PurePath:
    """Pydra task for creating a binary mask Nifti1Image from the list of tissues.

    .. note::
        Please refer to the documentation of function
        `clinica.pipelines.pet_volume.pet_volume_utils.create_binary_mask`.
    """
    from clinica.pipelines.pet_volume.pet_volume_utils import create_binary_mask

    return create_binary_mask(tissues, threshold)


@task
@annotate({"return": {"masked_image_path": PurePath}})
def apply_binary_mask_task(image: PurePath, binary_mask: PurePath) -> PurePath:
    """Pydra task for applying the provided `binary_mask` to the provided `image`.

    .. note::
        Please refer to the documentation of function
        `clinica.pipelines.pet_volume.pet_volume_utils.apply_binary_mask`.
    """
    from clinica.pipelines.pet_volume.pet_volume_utils import apply_binary_mask

    return apply_binary_mask(image, binary_mask)


@task
@annotate({"return": {"atlas_statistics": ty.List}})
def atlas_statistics_task(in_image: str, in_atlas_list: ty.List) -> ty.List:
    """Pydra task for generating regional measures from atlas_list in TSV files.

    .. note::
        Please refer to the documentation of function
        `clinica.pipelines.pet_volume.pet_volume_utils.atlas_statistics`.
    """
    from clinica.pipelines.pet_volume.pet_volume_utils import atlas_statistics

    return atlas_statistics(in_image, in_atlas_list)


@task
@annotate({"return": {"out_mask": PurePath}})
def create_pvc_mask_task(tissues: ty.List) -> PurePath:
    """Pydra task for creating a pvc mask from tissue list.

    .. note::
        Please refer to the documentation of function
        `clinica.pipelines.pet_volume.pet_volume_utils.create_pvc_mask`.
    """
    from clinica.pipelines.pet_volume.pet_volume_utils import create_pvc_mask

    return create_pvc_mask(tissues)


@task
@annotate({"return": {"pet_pvc_path": str}})
def pet_pvc_name_task(pet_image: PurePath, pvc_method: str) -> str:
    """Pydra task for building the name for the PET PVC interface.

    .. note::
        Please refer to the documentation of function
        `clinica.pipelines.pet_volume.pet_volume_utils.pet_pvc_name`.
    """
    from clinica.pipelines.pet_volume.pet_volume_utils import pet_pvc_name

    return pet_pvc_name(pet_image, pvc_method)


@task
@annotate({"return": {"psf_x": int, "psf_y": int, "psf_z": int}})
def get_psf_task(
    pvc_psf_tsv: PurePath,
    pet_filename: PurePath,
    pet_tracer: str,
) -> ty.Tuple[int]:
    """Returns the tuple (psf_x, psf_y, psf_z) for the subject at the
    session currently treated by the workflow.

    .. note::
        This function uses the name of the pet scan file to retrieve
        the subject and session IDs which are needed by the function
        `read_psf_information`.

    Parameters
    ----------
    pvc_psf_tsv : PurePath
        Path to the custom pvc_psf TSV file.
        See docstring of function `read_psf_information` to get more
        details on this file format.

    pet_filename : PurePath
        Path to the PET image for the subject and session treated by
        the workflow. See note above for details.

    pet_tracer : str
        Label of the PET tracer for which to retrieve psf.

    Returns
    -------
    (psf_x, psf_y, psf_z) : Tuple[int]
        The length 3 tuple for the subject session.
    """
    import re

    from clinica.utils.pet import read_psf_information

    m = re.search(r"(sub-[a-zA-Z0-9]+)_(ses-[a-zA-Z0-9]+)", str(pet_filename.name))
    if not m:
        raise ValueError(
            f"Could not extract subject and session IDs from filename {pet_filename}"
        )
    sub = m.group(1)
    ses = m.group(2)
    psf_x, psf_y, psf_z = read_psf_information(pvc_psf_tsv, [sub], [ses], pet_tracer)[0]
    return psf_x, psf_y, psf_z



"""This module contains utilities for statistics.

Currently, it contains one function to generate TSV file containing mean map based on a parcellation.
"""


def statistics_on_atlas(in_normalized_map, in_atlas, out_file=None):
    """Compute statistics of a map on an atlas.

    Given an atlas image with a set of ROIs, this function computes the mean of
    a normalized map (e.g. GM segmentation, FA map from DTI, etc.) on each ROI.

    Args:
        in_normalized_map (str): File containing a scalar image registered
            on the atlas.
        in_atlas (:obj: AbstractClass): An atlas with a set of ROI. These ROI
            are used to compute statistics.
        out_file (Optional[str]): Name of the output file.

    Returns:
        out_file (str): TSV file containing the statistics (content of the
            columns: label, mean scalar, std of the scalar', number of voxels).
    """
    import os.path as op

    import nibabel as nib
    import numpy as np
    import pandas

    from clinica.utils.atlas import AtlasAbstract
    from clinica.utils.stream import cprint

    if not isinstance(in_atlas, AtlasAbstract):
        raise Exception("Atlas element must be an AtlasAbstract type")

    if not out_file:
        fname, ext = op.splitext(op.basename(in_normalized_map))
        if ext == ".gz":
            fname, _ = op.splitext(fname)
        out_file = op.abspath(f"{fname}_statistics_{in_atlas.get_name_atlas()}.tsv")

    atlas_labels = nib.load(in_atlas.get_atlas_labels())
    atlas_labels_data = atlas_labels.get_fdata(dtype="float32")

    img = nib.load(in_normalized_map)
    img_data = img.get_fdata(dtype="float32")

    atlas_correspondence = pandas.read_csv(in_atlas.get_tsv_roi(), sep="\t")
    label_name = list(atlas_correspondence.roi_name)
    label_value = list(
        atlas_correspondence.roi_value
    )  # TODO create roi_value column in lut_*.txt and remove irrelevant RGB information

    mean_signal_value = []
    for label in label_value:
        current_mask_label = atlas_labels_data == label
        masked_data = np.array(img_data, copy=True)
        masked_data[np.invert(current_mask_label)] = 0
        mean_signal_value.append(np.sum(masked_data) / np.sum(current_mask_label))

    try:
        data = pandas.DataFrame(
            {"label_name": label_name, "mean_scalar": mean_signal_value}
        )
        data.to_csv(out_file, sep="\t", index=True, encoding="utf-8")
    except Exception as e:
        cprint(msg=f"Impossible to save {out_file} with pandas", lvl="error")
        raise e

    return out_file



import clinica.pipelines.engine as cpe


class StatisticsVolume(cpe.Pipeline):
    """StatisticsVolume - Volume-based mass-univariate analysis with SPM.

    Returns:
        A clinica pipeline object containing the StatisticsVolume pipeline.
    """

    def check_pipeline_parameters(self):
        """Check pipeline parameters."""
        from clinica.utils.exceptions import ClinicaException
        from clinica.utils.group import check_group_label

        # Clinica compulsory parameters
        self.parameters.setdefault("group_label", None)
        check_group_label(self.parameters["group_label"])

        if "orig_input_data_volume" not in self.parameters.keys():
            raise KeyError(
                "Missing compulsory orig_input_data_volume key in pipeline parameter."
            )

        if "contrast" not in self.parameters.keys():
            raise KeyError("Missing compulsory contrast key in pipeline parameter.")

        # Optional parameters
        self.parameters.setdefault("group_label_dartel", "*")
        self.parameters.setdefault("full_width_at_half_maximum", 8)

        # Optional parameters for inputs from pet-volume pipeline
        self.parameters.setdefault("acq_label", None)
        self.parameters.setdefault("suvr_reference_region", None)
        self.parameters.setdefault("use_pvc_data", False)

        # Optional parameters for custom pipeline
        self.parameters.setdefault("measure_label", None)
        self.parameters.setdefault("custom_file", None)

        # Advanced parameters
        self.parameters.setdefault("cluster_threshold", 0.001)

        if (
            self.parameters["cluster_threshold"] < 0
            or self.parameters["cluster_threshold"] > 1
        ):
            raise ClinicaException(
                "Cluster threshold should be between 0 and 1 "
                "(given value: %s)." % self.parameters["cluster_threshold"]
            )

    def check_custom_dependencies(self):
        """Check dependencies that can not be listed in the `info.json` file."""

    def get_input_fields(self):
        """Specify the list of possible inputs of this pipeline.

        Returns:
            A list of (string) input fields name.
        """
        return ["input_files"]

    def get_output_fields(self):
        """Specify the list of possible outputs of this pipeline.

        Returns:
            A list of (string) output fields name.
        """
        return [
            "spmT_0001",
            "spmT_0002",
            "spm_figures",
            "variance_of_error",
            "resels_per_voxels",
            "mask",
            "regression_coeff",
            "contrast",
        ]

    def build_input_node(self):
        """Build and connect an input node to the pipeline."""
        import nipype.interfaces.utility as nutil
        import nipype.pipeline.engine as npe

        from clinica.utils.exceptions import ClinicaException
        from clinica.utils.input_files import (
            pet_volume_normalized_suvr_pet,
            t1_volume_template_tpm_in_mni,
        )
        from clinica.utils.inputs import clinica_file_reader
        from clinica.utils.stream import cprint
        from clinica.utils.ux import print_begin_image, print_images_to_process

        all_errors = []
        if self.parameters["orig_input_data_volume"] == "pet-volume":
            if not (
                self.parameters["acq_label"]
                and self.parameters["suvr_reference_region"]
            ):
                raise ValueError(
                    f"Missing value(s) in parameters from pet-volume pipeline. Given values:\n"
                    f"- acq_label: {self.parameters['acq_label']}\n"
                    f"- suvr_reference_region: {self.parameters['suvr_reference_region']}\n"
                    f"- use_pvc_data: {self.parameters['use_pvc_data']}\n"
                )

            self.parameters["measure_label"] = self.parameters["acq_label"]
            information_dict = pet_volume_normalized_suvr_pet(
                acq_label=self.parameters["acq_label"],
                group_label=self.parameters["group_label_dartel"],
                suvr_reference_region=self.parameters["suvr_reference_region"],
                use_brainmasked_image=True,
                use_pvc_data=self.parameters["use_pvc_data"],
                fwhm=self.parameters["full_width_at_half_maximum"],
            )
        elif self.parameters["orig_input_data_volume"] == "t1-volume":
            self.parameters["measure_label"] = "graymatter"
            information_dict = t1_volume_template_tpm_in_mni(
                self.parameters["group_label_dartel"], 1, True
            )

        elif self.parameters["orig_input_data_volume"] == "custom-pipeline":
            if not self.parameters["custom_file"]:
                raise ClinicaException(
                    "Custom pipeline was selected but no 'custom_file' was specified."
                )
            # If custom file are grabbed, information of fwhm is irrelevant and should not appear on final filenames
            self.parameters["full_width_at_half_maximum"] = None
            information_dict = {
                "pattern": self.parameters["custom_file"],
                "description": "custom file provided by user",
            }
        else:
            raise ValueError(
                f"Input data {self.parameters['orig_input_data_volume']} unknown."
            )

        try:
            input_files, _ = clinica_file_reader(
                self.subjects, self.sessions, self.caps_directory, information_dict
            )
        except ClinicaException as e:
            all_errors.append(e)

        if len(all_errors) > 0:
            error_message = "Clinica faced errors while trying to read files in your CAPS directories.\n"
            for msg in all_errors:
                error_message += str(msg)
            raise ClinicaException(error_message)

        read_parameters_node = npe.Node(
            name="LoadingCLIArguments",
            interface=nutil.IdentityInterface(
                fields=self.get_input_fields(), mandatory_inputs=True
            ),
            synchronize=True,
        )
        read_parameters_node.inputs.input_files = input_files

        if len(self.subjects):
            print_images_to_process(self.subjects, self.sessions)
            cprint(
                "The pipeline will last a few minutes. Images generated by SPM will popup during the pipeline."
            )
            print_begin_image(f"group-{self.parameters['group_label']}")

        self.connect(
            [(read_parameters_node, self.input_node, [("input_files", "input_files")])]
        )

    def build_output_node(self):
        """Build and connect an output node to the pipeline."""
        from os.path import join, pardir

        import nipype.interfaces.io as nio
        import nipype.pipeline.engine as npe

        relative_path = join(
            "groups",
            "group-" + self.parameters["group_label"],
            "statistics_volume",
            "group_comparison_measure-" + self.parameters["measure_label"],
        )

        datasink = npe.Node(nio.DataSink(), name="sinker")
        datasink.inputs.base_directory = join(self.caps_directory, relative_path)

        datasink.inputs.parameterization = True
        if self.parameters["full_width_at_half_maximum"]:
            datasink.inputs.regexp_substitutions = [
                # t-stat map
                (
                    join(self.caps_directory, relative_path)
                    + r"/spm_results_analysis_./(.*)",
                    join(self.caps_directory, relative_path) + r"/\1",
                ),
                # contrasts
                (
                    join(self.caps_directory, relative_path) + r"/contrasts/(.*)",
                    join(self.caps_directory, relative_path) + r"/\1",
                ),
                # resels per voxels
                (
                    join(self.caps_directory, relative_path)
                    + "/resels_per_voxels/resels_per_voxel.nii",
                    join(self.caps_directory, relative_path)
                    + "/group-"
                    + self.parameters["group_label"]
                    + "_RPV.nii",
                ),
                # mask
                (
                    join(self.caps_directory, relative_path)
                    + "/mask/included_voxel_mask.nii",
                    join(self.caps_directory, relative_path)
                    + "/group-"
                    + self.parameters["group_label"]
                    + "_mask.nii",
                ),
                # variance of error
                (
                    join(self.caps_directory, relative_path)
                    + "/variance_of_error/(.*)",
                    join(self.caps_directory, relative_path) + r"/\1",
                ),
                # tsv file
                (
                    join(self.caps_directory, relative_path) + r"/tsv_file/.*",
                    join(self.caps_directory, relative_path)
                    + "/"
                    + pardir
                    + "/group-"
                    + self.parameters["group_label"]
                    + "_participants.tsv",
                ),
                # report (figures)
                (
                    join(self.caps_directory, relative_path) + r"/figures/(.*)",
                    join(self.caps_directory, relative_path) + r"/\1",
                ),
                # regression coefficient
                (
                    join(self.caps_directory, relative_path)
                    + r"/regression_coeff/(.*).nii",
                    join(self.caps_directory, relative_path)
                    + "/group-"
                    + self.parameters["group_label"]
                    + r"_covariate-\1"
                    + "_measure-"
                    + self.parameters["measure_label"]
                    + "_fwhm-"
                    + str(self.parameters["full_width_at_half_maximum"])
                    + "_regressionCoefficient.nii",
                ),
            ]

        datasink.inputs.tsv_file = self.tsv_file

        # fmt: off
        self.connect(
            [
                (self.output_node, datasink, [("spmT_0001", "spm_results_analysis_1")]),
                (self.output_node, datasink, [("spmT_0002", "spm_results_analysis_2")]),
                (self.output_node, datasink, [("spm_figures", "figures")]),
                (self.output_node, datasink, [("variance_of_error", "variance_of_error")]),
                (self.output_node, datasink, [("resels_per_voxels", "resels_per_voxels")]),
                (self.output_node, datasink, [("mask", "mask")]),
                (self.output_node, datasink, [("regression_coeff", "regression_coeff")]),
                (self.output_node, datasink, [("contrasts", "contrasts")]),
            ]
        )
        # fmt: on

    def build_core_nodes(self):
        """Build and connect the core nodes of the pipeline."""
        from os.path import dirname, join

        import nipype.interfaces.utility as nutil
        import nipype.pipeline.engine as npe

        import clinica.pipelines.statistics_volume.statistics_volume_utils as utils
        from clinica.utils.filemanip import unzip_nii

        # SPM cannot handle zipped files
        unzip_node = npe.Node(
            nutil.Function(
                input_names=["in_file"],
                output_names=["output_files"],
                function=unzip_nii,
            ),
            name="unzip_node",
        )

        # Get indexes of the 2 groups, based on the contrast column of the tsv file
        get_groups = npe.Node(
            nutil.Function(
                input_names=["tsv", "contrast"],
                output_names=["idx_group1", "idx_group2", "class_names"],
                function=utils.get_group_1_and_2,
            ),
            name="get_groups",
        )

        get_groups.inputs.contrast = self.parameters["contrast"]
        get_groups.inputs.tsv = self.tsv_file

        # Run SPM nodes are all a copy of a generic SPM script launcher
        run_spm_script_node = npe.Node(
            nutil.Function(
                input_names=["m_file"],
                output_names=["spm_mat"],
                function=utils.run_m_script,
            ),
            name="run_spm_script_node",
        )

        run_spm_model_creation = run_spm_script_node.clone(
            name="run_spm_model_creation"
        )
        run_spm_model_estimation = run_spm_script_node.clone(
            name="run_spm_model_estimation"
        )
        run_spm_model_contrast = run_spm_script_node.clone(
            name="run_spm_model_contrast"
        )
        run_spm_model_result_no_correction = run_spm_script_node.clone(
            name="run_spm_model_result_no_correction"
        )

        # All the following node are creating the correct (.m) script for the different SPM steps
        # 1. Model creation
        # 2. Model estimation
        # 3. Creation of contrast with covariates
        # 4. Creation of results

        # 1. Model creation
        # We use overwrite option to be sure this node is always run so that it can delete the output dir if it
        # already exists (this may cause error in output files otherwise)
        model_creation = npe.Node(
            nutil.Function(
                input_names=[
                    "tsv",
                    "contrast",
                    "idx_group1",
                    "idx_group2",
                    "file_list",
                    "template_file",
                ],
                output_names=["script_file", "covariates"],
                function=utils.model_creation,
            ),
            name="model_creation",
            overwrite=True,
        )
        model_creation.inputs.tsv = self.tsv_file
        model_creation.inputs.contrast = self.parameters["contrast"]
        model_creation.inputs.template_file = join(
            dirname(__file__), "template_model_creation.m"
        )

        # 2. Model estimation
        model_estimation = npe.Node(
            nutil.Function(
                input_names=["mat_file", "template_file"],
                output_names=["script_file"],
                function=utils.estimate,
            ),
            name="model_estimation",
        )
        model_estimation.inputs.template_file = join(
            dirname(__file__), "template_model_estimation.m"
        )

        # 3. Contrast
        model_contrast = npe.Node(
            nutil.Function(
                input_names=["mat_file", "template_file", "covariates", "class_names"],
                output_names=["script_file"],
                function=utils.contrast,
            ),
            name="model_contrast",
        )
        model_contrast.inputs.template_file = join(
            dirname(__file__), "template_model_contrast.m"
        )

        # 4. Results
        model_result_no_correction = npe.Node(
            nutil.Function(
                input_names=["mat_file", "template_file", "method", "threshold"],
                output_names=["script_file"],
                function=utils.results,
            ),
            name="model_result_no_correction",
        )
        model_result_no_correction.inputs.template_file = join(
            dirname(__file__), "template_model_results.m"
        )

        model_result_no_correction.inputs.method = "none"
        model_result_no_correction.inputs.threshold = self.parameters[
            "cluster_threshold"
        ]

        # Print result to txt file if spm

        # Export results to output node
        read_output_node = npe.Node(
            nutil.Function(
                input_names=[
                    "spm_mat",
                    "class_names",
                    "covariates",
                    "group_label",
                    "fwhm",
                    "measure",
                ],
                output_names=[
                    "spmT_0001",
                    "spmT_0002",
                    "spm_figures",
                    "variance_of_error",
                    "resels_per_voxels",
                    "mask",
                    "regression_coeff",
                    "contrasts",
                ],
                function=utils.read_output,
            ),
            name="read_output_node",
        )
        read_output_node.inputs.group_label = self.parameters["group_label"]
        read_output_node.inputs.fwhm = self.parameters["full_width_at_half_maximum"]
        read_output_node.inputs.measure = self.parameters["measure_label"]

        # Connection
        # ==========
        # fmt: off
        self.connect(
            [
                (self.input_node, unzip_node, [("input_files", "in_file")]),
                (unzip_node, model_creation, [("output_files", "file_list")]),
                (get_groups, model_creation, [("idx_group1", "idx_group1")]),
                (get_groups, model_creation, [("idx_group2", "idx_group2")]),
                (model_creation, run_spm_model_creation, [("script_file", "m_file")]),
                (run_spm_model_creation, model_estimation, [("spm_mat", "mat_file")]),
                (model_estimation, run_spm_model_estimation, [("script_file", "m_file")]),
                (get_groups, model_contrast, [("class_names", "class_names")]),
                (run_spm_model_estimation, model_contrast, [("spm_mat", "mat_file")]),
                (model_creation, model_contrast, [("covariates", "covariates")]),
                (model_contrast, run_spm_model_contrast, [("script_file", "m_file")]),
                (run_spm_model_contrast, model_result_no_correction, [("spm_mat", "mat_file")]),
                (model_result_no_correction, run_spm_model_result_no_correction, [("script_file", "m_file")]),
                (run_spm_model_result_no_correction, read_output_node, [("spm_mat", "spm_mat")]),
                (get_groups, read_output_node, [("class_names", "class_names")]),
                (model_creation, read_output_node, [("covariates", "covariates")]),
                (read_output_node, self.output_node, [("spmT_0001", "spmT_0001")]),
                (read_output_node, self.output_node, [("spmT_0002", "spmT_0002")]),
                (read_output_node, self.output_node, [("spm_figures", "spm_figures")]),
                (read_output_node, self.output_node, [("variance_of_error", "variance_of_error")]),
                (read_output_node, self.output_node, [("resels_per_voxels", "resels_per_voxels")]),
                (read_output_node, self.output_node, [("mask", "mask")]),
                (read_output_node, self.output_node, [("regression_coeff", "regression_coeff")]),
                (read_output_node, self.output_node, [("contrasts", "contrasts")]),
            ]
        )
        # fmt: on



def sort_session_list(session_list):
    session_idx = [int(session[5:]) for session in session_list]
    session_idx.sort()
    session_id_list = []
    for session in session_idx:
        if session < 10:
            session_id_list.append(f"ses-M0{session}")
        else:
            session_id_list.append(f"ses-M{session}")

    return session_id_list


def print_statistics(summary_file, num_subjs, ses_aval, mmt):
    """Write statistics file.

    Print to a given input file statistics about missing files and modalities
    in a dataset.  This metod takes in input a MissingModsTracker object (mmt)
    that contains the number of missing modalities for each session and the
    number of missing sessions for each subject.

    Args:
        summary_file: path of the output file where write.
        num_subjs: number of subjects.
        ses_aval: list of sessions available.
        mmt: object MissingModsTracker
    """
    missing_list = mmt.get_missing_list()
    ses_aval = sort_session_list(ses_aval)
    summary_file.write("**********************************************\n")
    summary_file.write(f"Number of subjects converted: {num_subjs}\n")
    summary_file.write(f"Sessions available: {ses_aval}\n")

    ses_aval = sort_session_list(ses_aval)

    for ses in ses_aval:
        ses_miss = missing_list[ses]["session"]
        ses_found = num_subjs - ses_miss
        perc_ses_found = ses_found * 100 / num_subjs
        summary_file.write(
            f"Number of sessions {ses} found: {ses_found} ({perc_ses_found}%)\n"
        )

    summary_file.write("**********************************************\n\n")
    summary_file.write("Number of missing modalities for each session:\n")

    for ses in ses_aval:
        summary_file.write("\n" + ses + "\n")
        for mod in missing_list[ses]:
            if mod != "session":
                num_miss_mod = missing_list[ses][mod]
                percentage_missing = round(
                    (
                        num_miss_mod
                        * 100
                        / float(num_subjs - missing_list[ses]["session"])
                    ),
                    2,
                )
                summary_file.write(f"{mod}: {num_miss_mod} ({percentage_missing}%) \n")


def print_longitudinal_analysis(
    summary_file, bids_dir, out_dir, ses_aval, out_file_name
):
    """Print to a given input file statistics about the present modalities and diagnoses in a dataset for each session.

    Args:
        summary_file: path of the output file where write.
        bids_dir: path to the BIDS directory
        out_dir: output_dir of the check-missing-modality pipeline.
        ses_aval: list of sessions available.
        out_file_name: string that replace the default prefix ('missing_mods_') in the name of all the output files
    """
    from os import path

    import pandas as pd

    ses_aval = sort_session_list(ses_aval)

    summary_file.write("**********************************************\n\n")
    summary_file.write("Number of present diagnoses and modalities for each session:\n")

    for ses in ses_aval:
        ses_df = pd.read_csv(
            path.join(out_dir, out_file_name + ses + ".tsv"), sep="\t"
        ).set_index("participant_id")
        mods_avail = ses_df.columns.values
        mod_dict = dict()
        for mod in mods_avail:
            diagnosis_dict = dict()
            subjects_avail = ses_df[ses_df[mod] == 1].index.values
            for subject in subjects_avail:
                subj_tsv_path = path.join(bids_dir, subject, f"{subject}_sessions.tsv")
                subj_df = pd.read_csv(subj_tsv_path, sep="\t").set_index("session_id")
                if (
                    ses in subj_df.index.values
                    and "diagnosis" in subj_df.columns.values
                ):
                    diagnosis = subj_df.loc[ses, "diagnosis"]
                    if isinstance(diagnosis, str):
                        increment_dict(diagnosis_dict, diagnosis)
                    else:
                        increment_dict(diagnosis_dict, "n/a")
                else:
                    increment_dict(diagnosis_dict, "missing")

            mod_dict[mod] = diagnosis_dict

        summary_file.write(f"{ses}\n")
        print_table(summary_file, mod_dict)
        summary_file.write("\n\n")


def increment_dict(dictionnary, key):
    if key not in dictionnary:
        dictionnary[key] = 1
    else:
        dictionnary[key] += 1


def print_table(summary_file, double_dict):

    # Find all keys at the second level of the dictionnary
    diagnoses = set()
    mods = double_dict.keys()
    for mod in mods:
        diagnoses = diagnoses | set(double_dict[mod].keys())

    diagnoses = list(diagnoses)
    diagnoses.sort()

    summary_file.write("\t")
    for diag in diagnoses:
        summary_file.write(f"\t| {diag}")
    summary_file.write("\n" + "-" * 8 * (len(diagnoses) + 2) + "\n")

    for mod in double_dict.keys():
        summary_file.write(f"{mod}")
        if len(mod) < 8:
            summary_file.write("\t")
        for diag in diagnoses:
            if diag not in double_dict[mod]:
                summary_file.write("\t| 0")
            else:
                summary_file.write(f"\t| {double_dict[mod][diag]}")
        summary_file.write("\n")


def has_one_index(index_list):
    if len(index_list) == 1:
        return index_list[0]
    if len(index_list) == 0:
        return -1
    if len(index_list) > 1:
        raise ValueError("Multiple indexes found")


class MissingModsTracker:
    """Class used for tracking the number of missing modalities in a database."""

    def __init__(self, ses, mod_list=None):
        self.missing = {}
        self.ses = ses
        if mod_list:
            for s in ses:
                self.missing.update({s: {"session": 0}})
                for mod in mod_list:
                    self.missing[s].update({mod: 0})
        else:
            for s in ses:
                self.missing.update(
                    {
                        s: {
                            "session": 0,
                            "dwi": 0,
                            "func": 0,
                            "fieldmap": 0,
                            "flair": 0,
                            "t1w": 0,
                        }
                    }
                )

    def add_missing_mod(self, ses, mod):
        """Increase the number of missing files for the input modality.

        Args:
            ses: name of the session
            mod: modality missing
        """
        self.missing[ses][mod] += 1

    def increase_missing_ses(self, ses):
        self.missing[ses]["session"] += 1

    def get_missing_list(self):
        """Return the hash map mods_missing.

        Returns:
             The hash map containing the list of missing files.
        """
        return self.missing



def eddy_fsl_pipeline(low_bval, use_cuda, initrand, name="eddy_fsl"):
    """Use FSL eddy for head motion correction and eddy current distortion correction."""
    import nipype.interfaces.utility as niu
    import nipype.pipeline.engine as pe
    from nipype.interfaces.fsl.epi import Eddy

    from clinica.utils.dwi import generate_acq_file, generate_index_file

    inputnode = pe.Node(
        niu.IdentityInterface(
            fields=[
                "in_file",
                "in_bvec",
                "in_bval",
                "in_mask",
                "ref_b0",
                "total_readout_time",
                "phase_encoding_direction",
            ]
        ),
        name="inputnode",
    )

    generate_acq = pe.Node(
        niu.Function(
            input_names=[
                "in_dwi",
                "fsl_phase_encoding_direction",
                "total_readout_time",
            ],
            output_names=["out_file"],
            function=generate_acq_file,
        ),
        name="generate_acq",
    )

    generate_index = pe.Node(
        niu.Function(
            input_names=["in_bval", "low_bval"],
            output_names=["out_file"],
            function=generate_index_file,
        ),
        name="generate_index",
    )
    generate_index.inputs.low_bval = low_bval

    eddy = pe.Node(interface=Eddy(), name="eddy_fsl")
    eddy.inputs.repol = True
    eddy.inputs.use_cuda = use_cuda
    eddy.inputs.initrand = initrand

    outputnode = pe.Node(
        niu.IdentityInterface(
            fields=["out_parameter", "out_corrected", "out_rotated_bvecs"]
        ),
        name="outputnode",
    )

    wf = pe.Workflow(name=name)
    # fmt: off
    wf.connect(
        [
            (inputnode, generate_acq, [('in_file', 'in_dwi')]),
            (inputnode, generate_acq, [('total_readout_time', 'total_readout_time')]),
            (inputnode, generate_acq, [('phase_encoding_direction', 'fsl_phase_encoding_direction')]),

            (inputnode, generate_index, [('in_bval', 'in_bval')]),

            (inputnode, eddy, [('in_bvec', 'in_bvec')]),
            (inputnode, eddy, [('in_bval', 'in_bval')]),
            (inputnode, eddy, [('in_file', 'in_file')]),
            (inputnode, eddy, [('in_mask', 'in_mask')]),
            (generate_acq, eddy, [('out_file', 'in_acqp')]),
            (generate_index, eddy, [('out_file', 'in_index')]),

            (eddy, outputnode, [('out_parameter', 'out_parameter')]),
            (eddy, outputnode, [('out_corrected', 'out_corrected')]),
            (eddy, outputnode, [('out_rotated_bvecs', 'out_rotated_bvecs')])
        ]
    )
    # fmt: on
    return wf


def epi_pipeline(
    base_dir: str,
    delete_cache: bool = False,
    name="susceptibility_distortion_correction_using_t1",
):
    """Perform EPI correction.

    This workflow allows to correct for echo-planar induced susceptibility artifacts without fieldmap
    (e.g. ADNI Database) by elastically register DWIs to their respective baseline T1-weighted
    structural scans using an inverse consistent registration algorithm with a mutual information cost
    function (SyN algorithm). This workflow allows also a coregistration of DWIs with their respective
    baseline T1-weighted structural scans in order to latter combine tracks and cortex parcellation.

    Parameters
    ----------
    base_dir: str
        Working directory, which contains all of the intermediary data generated.

    delete_cache: bool
        If True, part of the temporary data is automatically deleted after usage.

    name: str, optional
        Name of the pipeline.

    Warnings:
        This workflow rotates the b-vectors.

    Notes:
        Nir et al. (2015): Connectivity network measures predict volumetric atrophy in mild cognitive impairment
        Leow et al. (2007): Statistical Properties of Jacobian Maps and the Realization of
        Unbiased Large Deformation Nonlinear Image Registration
    """
    import nipype.interfaces.ants as ants
    import nipype.interfaces.c3 as c3
    import nipype.interfaces.fsl as fsl
    import nipype.interfaces.utility as niu
    import nipype.pipeline.engine as pe

    from .dwi_preprocessing_using_t1_utils import (
        ants_apply_transforms,
        change_itk_transform_type,
        delete_temp_dirs,
        expend_matrix_list,
        rotate_bvecs,
    )

    inputnode = pe.Node(
        niu.IdentityInterface(fields=["T1", "DWI", "bvec"]), name="inputnode"
    )

    split = pe.Node(fsl.Split(dimension="t"), name="SplitDWIs")
    pick_ref = pe.Node(niu.Select(), name="Pick_b0")
    pick_ref.inputs.index = [0]

    flirt_b0_2_t1 = pe.Node(interface=fsl.FLIRT(dof=6), name="flirt_B0_2_T1")
    flirt_b0_2_t1.inputs.interp = "spline"
    flirt_b0_2_t1.inputs.cost = "normmi"
    flirt_b0_2_t1.inputs.cost_func = "normmi"

    apply_xfm = pe.Node(interface=fsl.preprocess.ApplyXFM(), name="apply_xfm")
    apply_xfm.inputs.apply_xfm = True

    expend_matrix = pe.Node(
        interface=niu.Function(
            input_names=["in_matrix", "in_bvec"],
            output_names=["out_matrix_list"],
            function=expend_matrix_list,
        ),
        name="expend_matrix",
    )

    rot_bvec = pe.Node(
        niu.Function(
            input_names=["in_matrix", "in_bvec"],
            output_names=["out_file"],
            function=rotate_bvecs,
        ),
        name="Rotate_Bvec",
    )

    ants_registration = pe.Node(
        interface=ants.registration.RegistrationSynQuick(
            transform_type="br", dimension=3
        ),
        name="antsRegistrationSyNQuick",
    )

    c3d_flirt2ants = pe.Node(c3.C3dAffineTool(), name="fsl_reg_2_itk")
    c3d_flirt2ants.inputs.itk_transform = True
    c3d_flirt2ants.inputs.fsl2ras = True

    change_transform = pe.Node(
        niu.Function(
            input_names=["input_affine_file"],
            output_names=["updated_affine_file"],
            function=change_itk_transform_type,
        ),
        name="change_transform_type",
    )

    merge_transform = pe.Node(niu.Merge(3), name="MergeTransforms")

    # the nipype function is not used since the results it gives are not as good as the ones we get by using the command directly.
    apply_transform_image = pe.MapNode(
        interface=niu.Function(
            input_names=[
                "fixed_image",
                "moving_image",
                "transforms",
                "warped_image",
                "output_warped_image",
            ],
            output_names=["warped_image"],
            function=ants_apply_transforms,
        ),
        iterfield=["moving_image"],
        name="warp_image",
    )
    apply_transform_image.inputs.warped_image = "out_warped.nii.gz"
    apply_transform_image.inputs.output_warped_image = True

    apply_transform_field = pe.MapNode(
        interface=niu.Function(
            input_names=[
                "fixed_image",
                "moving_image",
                "transforms",
                "warped_image",
                "output_warped_image",
            ],
            output_names=["warped_image"],
            function=ants_apply_transforms,
        ),
        iterfield=["moving_image"],
        name="warp_field",
    )
    apply_transform_field.inputs.warped_image = "out_warped_field.nii.gz"
    apply_transform_field.inputs.output_warped_image = False

    jacobian = pe.MapNode(
        interface=ants.CreateJacobianDeterminantImage(),
        iterfield=["deformationField"],
        name="jacobian",
    )

    jacobian.inputs.imageDimension = 3
    jacobian.inputs.outputImage = "Jacobian_image.nii.gz"

    jacmult = pe.MapNode(
        fsl.MultiImageMaths(op_string="-mul %s"),
        iterfield=["in_file", "operand_files"],
        name="ModulateDWIs",
    )

    thres = pe.MapNode(
        fsl.Threshold(thresh=0.0), iterfield=["in_file"], name="RemoveNegative"
    )

    merge = pe.Node(fsl.Merge(dimension="t"), name="MergeDWIs")
    # Delete the temporary directory that takes too much place

    delete_warp_field_tmp = pe.Node(
        name="deletewarpfieldtmp",
        interface=niu.Function(
            inputs=["checkpoint", "dir_to_del", "base_dir"],
            function=delete_temp_dirs,
        ),
    )
    delete_warp_field_tmp.inputs.base_dir = base_dir
    delete_warp_field_tmp.inputs.dir_to_del = [
        apply_transform_field.name,
        jacobian.name,
        jacmult.name,
        thres.name,
        apply_transform_image.name,
    ]

    outputnode = pe.Node(
        niu.IdentityInterface(
            fields=[
                "DWI_2_T1_Coregistration_matrix",
                "epi_correction_deformation_field",
                "epi_correction_affine_transform",
                "epi_correction_image_warped",
                "DWIs_epicorrected",
                "warp_epi",
                "out_bvec",
            ]
        ),
        name="outputnode",
    )

    wf = pe.Workflow(name="epi_pipeline")
    # fmt: off
    wf.connect(
        [
            (inputnode, split, [("DWI", "in_file")]),
            (split, pick_ref, [("out_files", "inlist")]),
            (pick_ref, flirt_b0_2_t1, [("out", "in_file")]),
            (inputnode, flirt_b0_2_t1, [("T1", "reference")]),
            (inputnode, rot_bvec, [("bvec", "in_bvec")]),
            (flirt_b0_2_t1, expend_matrix, [("out_matrix_file", "in_matrix")]),
            (inputnode, expend_matrix, [("bvec", "in_bvec")]),
            (expend_matrix, rot_bvec, [("out_matrix_list", "in_matrix")]),
            (inputnode, ants_registration, [("T1", "fixed_image")]),
            (flirt_b0_2_t1, ants_registration, [("out_file", "moving_image")]),
            (inputnode, c3d_flirt2ants, [("T1", "reference_file")]),
            (pick_ref, c3d_flirt2ants, [("out", "source_file")]),
            (flirt_b0_2_t1, c3d_flirt2ants, [("out_matrix_file", "transform_file")]),
            (c3d_flirt2ants, change_transform, [("itk_transform", "input_affine_file")]),
            (change_transform, merge_transform, [("updated_affine_file", "in1")]),
            (ants_registration, merge_transform, [("out_matrix", "in2")]),
            (ants_registration, merge_transform, [("forward_warp_field", "in3")]),
            (inputnode, apply_transform_image, [("T1", "fixed_image")]),
            (split, apply_transform_image, [("out_files", "moving_image")]),
            (merge_transform, apply_transform_image, [("out", "transforms")]),

            (inputnode, apply_transform_field, [("T1", "fixed_image")]),
            (split, apply_transform_field, [("out_files", "moving_image")]),
            (merge_transform, apply_transform_field, [("out", "transforms")]),

            (apply_transform_field, jacobian, [("warped_image", "deformationField")]),
            (apply_transform_image, jacmult, [("warped_image", "operand_files")]),
            (jacobian, jacmult, [("jacobian_image", "in_file")]),
            (jacmult, thres, [("out_file", "in_file")]),
            (thres, merge, [("out_file", "in_files")]),
            
            (merge, outputnode, [("merged_file", "DWIs_epicorrected")]),
            (flirt_b0_2_t1, outputnode, [("out_matrix_file", "DWI_2_T1_Coregistration_matrix")]),
            (ants_registration, outputnode, [("forward_warp_field", "epi_correction_deformation_field"),
                                             ("out_matrix", "epi_correction_affine_transform"),
                                             ("warped_image", "epi_correction_image_warped")]),
            (merge_transform, outputnode, [("out", "warp_epi")]),
            (rot_bvec, outputnode, [("out_file", "out_bvec")]),

            
            
        ]
    )
    if delete_cache:
        wf.connect(
            [
                (merge, delete_warp_field_tmp, [("merged_file", "checkpoint")])
            ]
        )
    # fmt: on
    return wf


def b0_flirt_pipeline(num_b0s, name="b0_coregistration"):
    """Rigid registration of the B0 dataset onto the first volume.

    Rigid registration is achieved using FLIRT and the normalized correlation.

    Args:
        num_b0s (int): Number of the B0 volumes in the dataset.
        name (str): Name of the workflow.

    Inputnode:
        in_file(str): B0 dataset.

    Outputnode
        out_b0_reg(str): The set of B0 volumes registered to the first volume.
    """
    import nipype.interfaces.utility as niu
    import nipype.pipeline.engine as pe
    from nipype.interfaces import fsl

    from clinica.utils.dwi import merge_volumes_tdim

    inputnode = pe.Node(niu.IdentityInterface(fields=["in_file"]), name="inputnode")
    fslroi_ref = pe.Node(fsl.ExtractROI(args="0 1"), name="b0_reference")
    tsize = num_b0s - 1
    fslroi_moving = pe.Node(fsl.ExtractROI(args="1 " + str(tsize)), name="b0_moving")
    split_moving = pe.Node(fsl.Split(dimension="t"), name="split_b0_moving")

    bet_ref = pe.Node(fsl.BET(frac=0.3, mask=True, robust=True), name="bet_ref")

    dilate = pe.Node(
        fsl.maths.MathsCommand(nan2zeros=True, args="-kernel sphere 5 -dilM"),
        name="mask_dilate",
    )

    flirt = pe.MapNode(
        fsl.FLIRT(
            interp="spline",
            dof=6,
            bins=50,
            save_log=True,
            cost="corratio",
            cost_func="corratio",
            padding_size=10,
            searchr_x=[-4, 4],
            searchr_y=[-4, 4],
            searchr_z=[-4, 4],
            fine_search=1,
            coarse_search=10,
        ),
        name="b0_co_registration",
        iterfield=["in_file"],
    )

    merge = pe.Node(fsl.Merge(dimension="t"), name="merge_registered_b0s")
    thres = pe.MapNode(
        fsl.Threshold(thresh=0.0), iterfield=["in_file"], name="remove_negative"
    )
    insert_ref = pe.Node(
        niu.Function(
            input_names=["in_file1", "in_file2"],
            output_names=["out_file"],
            function=merge_volumes_tdim,
        ),
        name="concat_ref_moving",
    )

    outputnode = pe.Node(
        niu.IdentityInterface(fields=["out_file", "out_xfms"]), name="outputnode"
    )

    wf = pe.Workflow(name=name)
    # fmt: off
    wf.connect(
        [
            (inputnode, fslroi_ref, [("in_file", "in_file")]),
            (inputnode, fslroi_moving, [("in_file", "in_file")]),
            (fslroi_moving, split_moving, [("roi_file", "in_file")]),
            (fslroi_ref, bet_ref, [("roi_file", "in_file")]),
            (bet_ref, dilate, [("mask_file", "in_file")]),
            (dilate, flirt, [("out_file", "ref_weight"),
                             ("out_file", "in_weight")]),
            (fslroi_ref, flirt, [("roi_file", "reference")]),
            (split_moving, flirt, [("out_files", "in_file")]),
            (flirt, thres, [("out_file", "in_file")]),
            (thres, merge, [("out_file", "in_files")]),
            (merge, insert_ref, [("merged_file", "in_file2")]),
            (fslroi_ref, insert_ref, [("roi_file", "in_file1")]),
            (insert_ref, outputnode, [("out_file", "out_file")]),
            (flirt, outputnode, [("out_matrix_file", "out_xfms")])
        ]
    )
    # fmt: off

    return wf



from typing import Optional

import click

from clinica.pipelines import cli_param
from clinica.pipelines.cli import cli as run_cli

pipeline_name = "flair-linear"


@click.command(name=pipeline_name)
@cli_param.argument.bids_directory
@cli_param.argument.caps_directory
@cli_param.option_group.pipeline_specific_options
@cli_param.option_group.option(
    "-ui",
    "--uncropped_image",
    is_flag=True,
    help="Do not crop the image with template (cropped image are suggested for using with DL models)",
)
@cli_param.option_group.common_pipelines_options
@cli_param.option.subjects_sessions_tsv
@cli_param.option.working_directory
@cli_param.option.n_procs
def cli(
    bids_directory: str,
    caps_directory: str,
    uncropped_image: bool = False,
    subjects_sessions_tsv: Optional[str] = None,
    working_directory: Optional[str] = None,
    n_procs: Optional[int] = None,
) -> None:
    """Affine registration of Flair images to the MNI standard space.

    https://aramislab.paris.inria.fr/clinica/docs/public/latest/Pipelines/T1_Linear/
    """
    from networkx import Graph

    from clinica.utils.ux import print_end_pipeline

    from .anat_linear_pipeline import AnatLinear

    parameters = {"uncropped_image": uncropped_image}

    # Most of the time, you will want to instantiate your pipeline with a
    # BIDS and CAPS directory as inputs:
    pipeline = AnatLinear(
        bids_directory=bids_directory,
        caps_directory=caps_directory,
        tsv_file=subjects_sessions_tsv,
        base_dir=working_directory,
        parameters=parameters,
        name=pipeline_name,
    )

    exec_pipeline = (
        pipeline.run(plugin="MultiProc", plugin_args={"n_procs": n_procs})
        if n_procs
        else pipeline.run()
    )

    if isinstance(exec_pipeline, Graph):
        print_end_pipeline(
            pipeline_name, pipeline.base_dir, pipeline.base_dir_was_specified
        )


run_cli.add_command(cli)


if __name__ == "__main__":
    cli()



def atlas_statistics(in_image, atlas_list):
    """Generate regional measure from atlas_list in TSV files.

    For each atlas name provided it calculates for the input image the mean
    for each region in the atlas and saves it to a TSV file.

    Args:
        in_image: A Nifti image
        atlas_list: List of names of atlas to be applied

    Returns:
        List of paths to TSV files
    """
    from os.path import abspath, join

    from nipype.utils.filemanip import split_filename

    from clinica.utils.atlas import AtlasAbstract
    from clinica.utils.filemanip import get_subject_id
    from clinica.utils.statistics import statistics_on_atlas
    from clinica.utils.ux import print_end_image

    subject_id = get_subject_id(in_image)

    orig_dir, base, ext = split_filename(in_image)
    atlas_classes = AtlasAbstract.__subclasses__()
    atlas_statistics_list = []
    for atlas in atlas_list:
        for atlas_class in atlas_classes:
            if atlas_class.get_name_atlas() == atlas:
                out_atlas_statistics = abspath(
                    join(f"./{base}_space-{atlas}_map-graymatter_statistics.tsv")
                )
                statistics_on_atlas(in_image, atlas_class(), out_atlas_statistics)
                atlas_statistics_list.append(out_atlas_statistics)
    print_end_image(subject_id)
    return atlas_statistics_list
